/******************************************************************************
 * 	Copyright (c) 2008 Mark Hummel DBA Raquette Software.
 *		All rights reserved
 *
 *    This file contains source code written by Raquette Software, 
 *    68 Stewart Street, Franklin MA 02038. It may not be used without 
 *    express written permission. The expression of the information 
 *    contained herein is protected under federal copyright laws and 
 *    all copying without permission is prohibited and may be subject 
 *    to criminal penalties. The Author assumes no responsibility for 
 *    errors, omissions, or damages caused by the use of these programs 
 *    or from use of the information contained herein.
 *
 *-----------------------------------------------------------------------------
 */
/******************************************************************************
 *
 *
 *	   dump.cc
 *		- dump veriwell internal structures
 *
 ******************************************************************************
 */
						

#include <iostream>
#include "dump.h"

using namespace std;
using namespace veriwell;


unsigned tabs = 0;
void Tab() { for( int i=0; i<tabs; ++i ) cout << "  "; }

void DumpTree( tree n, bool chaseTree = true ) {
    if( !n ) {
	cout << "NULL";
	return;
    }
    switch( TREE_CODE(n) ) {
    case PARAM_DECL:
    case REG_VECTOR_DECL:
    case REG_SCALAR_DECL:
    case NET_VECTOR_DECL:
    case NET_SCALAR_DECL:
   	chaseTree = false; 
    }
    if( chaseTree && TREE_CHAIN(n) )  {
	cout << endl;
	Tab();
	cout  << "{"; 
	tabs++;
	for( tree c = n; c; c = TREE_CHAIN(c) ) {
		if( c != n ) {
		    cout << " -> ";
		}
		DumpTree( c, false );
	}
	cout << endl;
	tabs--;
	Tab();
	cout << "}";
	return;
    }
//    if( TREE_TYPE(n) == 'd' ) {
//        cout << tree_code_name[TREE_CODE(n)] << " " << DECL_NAME(n) << " ";
//	if( TREE_CHAIN(n) ) {
//	    cout << "->";
//	    DumpTree(TREE_CHAIN(n), false);
//	    cout << " ";
//	}
//	return;
//    }
	
    switch( TREE_CODE(n) ) {
    case IDENTIFIER_NODE:
	cout << IDENT(n);
	break;
    case WHILE_STMT:
	cout << endl;
	Tab();
        cout << tree_code_name[TREE_CODE(n)] << " "; 
	tabs++;
	DumpTree( STMT_OPERAND(n,0), false );
	tabs--;
	break;
    case IF_STMT:
	cout << endl;
	Tab();
        cout << tree_code_name[TREE_CODE(n)] << " "; 
	tabs++;
	     DumpTree( STMT_THEN(n) );
	cout << " ELSE ";
	     DumpTree( STMT_ELSE(n) );
	tabs--;
	break;
    case FORCE_STMT:
	cout << endl;
	Tab();
//        cout << STMT_SOURCE_FILE(n) << ":"; 
//        cout << STMT_SOURCE_LINE(n) << " "; 
        cout << tree_code_name[TREE_CODE(n)] << " "; 
	break;
    case CASEZ_STMT:
    case TASK_STMT:
    case SYSTASK_STMT:
    case ASSIGN_STMT:
    case ASSIGN_NONBLK_DELAY_STMT:
    case END_STMT:
    case NULL_STMT:
	cout << endl;
	Tab();
        cout << tree_code_name[TREE_CODE(n)] << " "; 
	break;
    case EVENT_STMT:
    case BEGIN_STMT:
    case DELAY_STMT:
	cout << endl;
	Tab();
        cout << tree_code_name[TREE_CODE(n)] << " "; 
	DumpTree( STMT_BODY(n) );
	break;
    case ALWAYS_BLOCK:
    case INITIAL_BLOCK:
	cout << endl;
	Tab();
        cout << tree_code_name[TREE_CODE(n)] << " "; 
	DumpTree( STMT_OPERAND(n,0), false );
	break;
    case BLOCK_DECL:
    case PARAM_DECL:
    case REG_VECTOR_DECL:
    case REG_SCALAR_DECL:
    case NET_VECTOR_DECL:
    case NET_SCALAR_DECL:
	cout << endl;
	Tab();
        cout << tree_code_name[TREE_CODE(n)] << " "; 
	DumpTree( DECL_NAME(n), false );
	break;
    case INSTANCE_NODE:
	cout << endl;
	Tab();
        cout << tree_code_name[TREE_CODE(n)] << " "; 
	DumpTree( INSTANCE_MODULE_NAME(n), false );     
	cout << " ";
	DumpTree( INSTANCE_NAME(n), false );     
	cout << " ";
	break;	
    case TREE_LIST:
        cout << endl; 
	Tab();
        cout << "("; 
	DumpTree( TREE_PURPOSE(n) );
        cout << ","; 
	if( TREE_VALUE(n) == (tree)ALWAYS_CODE ) {
	    cout << "ALWAYS_CODE";
	} else if( TREE_VALUE(n) == (tree)INITIAL_CODE ) {
	    cout << "INITIAL_CODE";
	} else {
	    DumpTree( TREE_VALUE(n) );
	}
        cout << ","; 
	DumpTree( TREE_3RD(n) );
	cout << ")";
	cout << endl;
	break;
    case TASK_BLOCK:
        cout << tree_code_name[TREE_CODE(n)] << " ";
	DumpTree( BLOCK_NAME(n), false ); 
	cout << " ";
	tabs++;
	cout << endl << "<<<";
	DumpTree( BLOCK_BODY(n) );
	cout << endl << ">>>";
	tabs--;
	break;
    case MODULE_BLOCK:
        cout << "module "<< MODULE_NAME(n) << " ";
	tabs++;
	Tab();
	cout << endl << "<<<";
	DumpTree( MODULE_PORT_LIST(n) );
	cout << endl << ">>>";
	Tab();
	cout << endl << "<<<";
	DumpTree( BLOCK_DECL(n) );
	cout << endl << ">>>";
	Tab();
	cout << endl << "<<<";
	DumpTree( BLOCK_BODY(n) );
	cout << endl << ">>>";
	Tab();
	cout << endl << "<<<";
	DumpTree( BLOCK_DOWN(n) );
	cout << endl << ">>>";
	tabs--;
        break;
    default:
        cout << "tree ??? " << n << " " << tree_code_name[TREE_CODE(n)] << endl;
    }

}

void DumpVeriwell()
{
    DumpTree( module_list );
}
