/*****************************************************************************
 * Copyright 1994-2005, Elliot Mednick and Mark Hummel
 * This file is part of Veriwell.
 *
 * Veriwell is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Veriwell is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Foobar; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *****************************************************************************/

/* EVAL.H - Macros and prototypes for Eval */

#ifndef EVAL_H
#define EVAL_H

#define R_MASK  MaskRight1(R_nbits)
#define R_NGROUPS  bits_to_groups(R_nbits) 
#define SET_R_NBITS(b) R_nbits = b; 				

namespace veriwell {
// public methods

/* This macro determines if the value point to by G is 0, 1, or X/Z */
/* (THIS MACRO ALSO APPEARS IN EXEC.C) */

#define DETERMINE_TRUE_COND(G)                          \
        {G = *--R;                                      \
        cond = ZERO;                                    \
        for (i = 0; i < R_NGROUPS; i++, (G)++) {        \
          if (BVAL (G)) {                               \
            cond = X;                                   \
            G = *R + R_NGROUPS;                         \
            break;                                      \
          }                                             \
          else if (AVAL (G))                            \
            cond = ONE;                                 \
        }                                               \
        if (BVAL (G) & R_MASK)                          \
          cond = X;                                     \
        else if (AVAL (G) & R_MASK)                     \
          cond = ONE;}

#define SET_STATE(G)                                    \
                {                                       \
                  if (DECL_STATE (lval) != X) {         \
                    if (BVAL (G) & AVAL (G))            \
                      SET_DECL_STATE (lval, X);         \
                    else if (DECL_STATE (lval) != Z) {  \
                      if (BVAL (G))                     \
                        SET_DECL_STATE (lval, Z);       \
                      else if (AVAL (G))                \
                        SET_DECL_STATE (lval, ONE);     \
                    }                                   \
                  }                                     \
                }

#define TEST(EXPR, G) eval (EXPR); DETERMINE_TRUE_COND (G)

extern Bit mask_right[];
inline Bit MaskRight1(unsigned i) { int j=i&NBITS_MASK;
				    return j==0 ? (Bit)-1 : (1<<j)-1; } 

void do_net_eval(tree node, int exclude_inout, tree exclude_tree);
extern Group **R;
extern nbits_t R_nbits;
inline void do_net_eval_fast(tree node)
{
    /*
     * accelerate single driver/single location WIRE case
     */
    tree source = NET_SOURCE(node);
    if( TREE_NBITS(node) <= 32 && TREE_TYPE(node) == NET_WIRE_TYPE && 
	source && NET_SOURCE(source) == NULL ) {
	SET_R_NBITS(TREE_NBITS(node));
        **R = *(DECL_STORAGE(source));
        *++R = *R+1;			// + 1;
	return;
    }
    do_net_eval(node, 0, NULL);
}
void eval(tree * pc);

/* Evaluate an expression to be equal to 1 */
inline int test1(tree * expr)
{
    int result = 0;
    Group *g;

    eval(expr);
    g = *--R;
    for (int i = 0; i < R_NGROUPS; i++, g++) {
        if(BVAL(g)) {
	   return 0;
        } else if (AVAL(g)) {
            result = 1;
        }
    }
    if(BVAL(g) & R_MASK) {
        return 0;
    } 
    if (AVAL(g) & R_MASK) {
	result = 1;
    }
    return result;
}
 /* Evaluate an expression;  return an integer non-zero if the top
  * of stack is non-zero.  Then pop the stack. 
  */

inline enum logical_value test(tree * expr)
{
    enum logical_value cond;
    ngroups_t i;
    Group *g;

    eval(expr);
    DETERMINE_TRUE_COND(g);
    return cond;
}


int part_lref(Group * g1, Group * g2, struct part_info *info);
ngroups_t part_rref(Group * g1, Group * g2, struct part_info *info);
Group *eval_(tree * pc, nbits_t * nbits);
void eval_1(tree decl);
delay_t eval_delay(tree delay_node, enum logical_value state);
int R_alloc(int number, int area);
nbits_t eval_range(tree * inst, char *ident);
Bit eval_const(tree * inst);
void eval_nbits(tree, nbits_t);
void init_eval();
double reg_to_real(Group * g, nbits_t nbits);
void real_to_reg(double x, Group * g, ngroups_t ngroups);
}
#endif				// EVAL_H
