/*****************************************************************************
 * Copyright 1994-2005, Elliot Mednick and Mark Hummel
 * This file is part of Veriwell.
 *
 * Veriwell is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Veriwell is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Foobar; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *****************************************************************************/

/* EXEC.H - Macros and prototypes for Exec */

#ifndef EXEC_H
#define EXEC_H

#include <time.h>
#include "flags.h"
#include "io.h"
#include "print.h"

namespace veriwell {
extern clock_t clock_pause;
extern int in_simulation;
extern int errorcount;	
extern int warningcount;	
extern int is_interactive;

void flush_files(void);
tree exec_(tree pc, int s_flag=0);
struct context_member *make_context(void);
void enter_context(struct SCB *scb, tree scope, tree pc);
tree exit_context(struct SCB *scb);
tree unwind_context(struct context_member *context);
void disable_context(struct context_member *context);
tree deassign(tree lval, enum tree_code code);
void do_net_assignment(tree pc, int force=0);
void initialize_cont_assignments(void);
void net_exec(tree pc);
void exec_history();
void exec_remove_history();
tree go_interactive(tree current_pc, tree next_pc, Group * g,
		    int skip, int val);
void handle_gate(tree gate);
void init_interactive(void);
void init_exec(void);

extern struct context_member *free_context;
inline void set_simulation_time(time64 * newtime_p)
{
    ASSIGNTIME64(&CurrentTime, newtime_p);
    if ((trace_flag || single_trace_flag) &&
	 CMPTIME64(&CurrentTime, &(SCB::ActiveSCB()->time))) {
	printf_V("SIMULATION TIME IS ");
	print_time(&CurrentTime);
	printf_V("\n");
    }
}

inline tree exit_context(SCB * scb)
{
    struct context_member *tmp;
    struct context_member *context = scb->context;

    scb->context = context->shadow;
    REMOVE_LIST(context);
    ADD_LIST(free_context, context);
    return context->pc;
}


/* Starting with the innermost context, unwind until past the given
   context.  Return the outermost PC. */

inline tree unwind_context(struct context_member * context)
{
    while (context->scb->context != context) {
	exit_context(context->scb);
	free_context->pc = NULL;
    }
    return exit_context(context->scb);
}

void eval(tree * pc);
void store(tree lval, tree pc);
inline void do_net_assignment_optimized(Marker* marker )
{
    tree pc = marker->expr.tree;

    /* 
     * if this is a force and activation is caused by
     * net or varible being changed, don't reevaluate
     * rval, used cached value.
     */
    tree* rvalCode = STMT_ASSIGN_RVAL_CODE(pc);
    eval(rvalCode);
    tree tmp = marker->tmp1;
    if( pc == tmp ) {
        store(STMT_ASSIGN_LVAL(pc), pc);
    }
    if( pc != tmp ) {
        store(tmp, tmp);
    }
}

inline void do_net_assignment(Marker* marker )
{
    tree pc = marker->expr.tree;
    int force = marker->flags & M_FORCE; 

    do_net_assignment(pc, force );
    /* 
     * optimize execution if possible
     */
    tree  lval = STMT_ASSIGN_LVAL(pc);
    if( !marker->tmp1 ) {
	marker->tmp1 =  pc;
	if( optimize && !force &&
 	    TREE_CODE(pc) != ASSIGN_PROC_STMT && 
	    TREE_CODE(pc) != FORCE_STMT ) { 
            if( !NET_ASSIGN_ATTR(lval) || 
		PORT_MIRRORED_ATTR(TREE_CHAIN(lval))) {
	        marker->flags = (enum marker_flags) 
					(M_OPTIMIZED | marker->flags);
	    } else {
                tree  net = TREE_CHAIN(lval);
                tree  source = NET_SOURCE(net);
	        if( !NO_OPTIMIZE_ATTR(net) && NET_SOURCE(source) == NULL &&
                    ((enum tree_type) TREE_TYPE(net) == NET_WIRE_TYPE ||
                    (enum tree_type) TREE_TYPE(net) == NET_TRI_TYPE ) ) {
	            marker->flags = (enum marker_flags) 
					(M_OPTIMIZED | marker->flags);
	            marker->tmp1 =  net;
	        }
	    }
	}
    }
}

}

#endif				// EXEC_H
