/*****************************************************************************
 * Copyright 1994-2005, Elliot Mednick and Mark Hummel
 * This file is part of Veriwell.
 *
 * Veriwell is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Veriwell is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Foobar; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *****************************************************************************/

/* FLAGS.C - Global storage for command line options */

#include "vtypes.h"
#include "tree.h"
#include "flags.h"

namespace veriwell {
/* For internal debugging: */
int program = 0;		/* displays program */
int debug = 0;			/* turns on internal trace */
#if VDEBUG != 0
int vdebug = 0;
#endif

/* Suppress warnings */

/* Enter trace mode upon entry.  Also used when $settrace is enabled. */
int trace_flag = 0;

/* Runtime state bits */
/* (trace_flag is in both flags.h and runtime.h) */
/* This is set if not trace nor single-step */
int normal_flag = 1;

/* set after ctl-c is pressed */
int break_flag = 0;

/* Set for either of the single-step commands: comma and semicolon */
int single_step_flag = 0;

/* set if comma (single-step with trace) */
int single_trace_flag = 0;

/* Capture all the user-defined plus args in a TREE_LIST */
union tree_node *plusargs = 0;

/* Set to 1 when exec should check for other flags in main loop */
int execAttention = 0;

/* Set to 1 dump runtime stats */
int printStats = 0;

enum tree_type default_net_type = NET_WIRE_TYPE;


int    globalArgc;
char** globalArgv;

 
/***************************************************************
 *
 *  init_flags
 *	- initialize all global and static variables in flags.c
 *
 ***************************************************************
 */

void init_flags()
{
    globalArgc = 0;
    globalArgv = 0;
    program = 0;
    debug = 0;
#if VDEBUG != 0
    vdebug = 0;
#endif
    trace_flag = 0;
    normal_flag = 1;
    execAttention = 0;
    printStats = 0;
    break_flag = 0;
    single_step_flag = 0;
    single_trace_flag = 0;
    plusargs = 0;
    default_net_type = NET_WIRE_TYPE;
}

}
