/*****************************************************************************
 * Copyright 1994-2005, Elliot Mednick and Mark Hummel
 * This file is part of Veriwell.
 *
 * Veriwell is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Veriwell is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Foobar; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *****************************************************************************/

/* IO.H - IO prototypes */

#ifndef IO_H
#define IO_H

#include <stdio.h>
#include <stdarg.h>
#include "vtypes.h"
#include "tree.h"


namespace veriwell {
// public methods

/*
 * LibPath_t
 * - structure to maintain linked list of library paths 
 */

typedef struct LibPath_t {
    struct LibPath_t *next;
    const char *path;
} LibPath_t;


#define NHANDLES 31

void printf_V_log(const char *, ...);
void printf_V(const char *, ...);
void printf_error_V(const char *, ...);
void printf_V_key(const char *, ...);
void fprintf_V(handle_t, const char *, ...);
void vprintf_V_log(const char *, va_list);
void vprintf_V(const char *, va_list);
void vfprintf_V(handle_t, const char *, va_list);
handle_t fopen_V(const char *);
handle_t check_handle(handle_t);
void fclose_V(handle_t);
void fflush_V(handle_t);
void error_with_file_and_line(const char*, lineno_t, const char*, const char*, const char*);
void error(const char*, const char*, const char*);
void warning_with_file_and_line(const char*, lineno_t, const char*, const char*, const char*);
void warning(const char*, const char*, const char*);
void sorry_with_file_and_line(const char*, lineno_t, const char*, const char*, const char*);
void sorry(const char*);
void sorry1(const char*, const char*);

FILE *fopenbuf(const char *, int);
char *xmalloc(unsigned);
char *xmalloc_aligned(unsigned);
char *xrealloc(char*, int);

void init_io(void);


class File;
extern File *fin;
extern const char* input_filename;
extern const char *filename;

extern FILE *log_file;
extern FILE *key_file;
extern char *log_file_name;
extern char *key_file_name;
extern int log_enable;
extern int log_available;
extern int key_enable;
extern int key_available;
extern char *ypath;
extern char *ylibext;
extern char *incdir;
extern int global_print_override;
extern char *global_print_buffer;
extern char *global_print_p;
extern LibPath_t *ypathList;

void fatal(const char *s, const char *v);
void runtime_error(tree node);

}

#endif				//IO_H


