/*****************************************************************************
 * Copyright 1994-2005, Elliot Mednick and Mark Hummel
 * This file is part of Veriwell.
 *
 * Veriwell is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Veriwell is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Foobar; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *****************************************************************************/

/* PASS2.C - Generate hierarchy, allocate variables, propigate parameters */

#define PASS2_C

#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <list>
#include "vtypes.h"
#include "tree.h"
#include "schedule.h"
#include "obstack.h"
#include "runtime.h"
#include "decl.h"
#include "scope.h"
#include "io.h"
#include "store.h"
#include "udp.h"
#include "pass3.h"
#include "eval.h"
#include "copy.h"
#include "check.h"
#include "pass2.h"
#include "flags.h"
#include "veriwellinterface.h"

using namespace std;

namespace veriwell {
static void parse_at_top_scope(tree scope);


#define obstack_chunk_alloc xmalloc
#define obstack_chunk_free free


/* This obstack contains pointers to expression/ref/decl nodes in the
   order that they should be evaluated (postfix). */
extern struct obstack inst_obstack;
extern struct obstack alt_inst_obstack;

/* This is some global information that needs to be kept around for
   when markers need to be generated for references inside event
   expressions.  Perhaps this is a grungy way of doing it.  */

extern struct Marker_info marker_info;

/* TREE_LIST of continuous assignments that are port connections */
tree port_connections = NULL_TREE;

/*
 * list of nets with the collapsed attribute set
 */
list<tree> collapsingNets;


/* Memory allocation routines for pass3_decl.  Allocate the memory and
   initialize to all Xs. */

Group *malloc_X(nbits_t nbits)
{
    ngroups_t ngroups = bits_to_groups(nbits);
    ngroups_t i;
    Group *tmp, *tmp1;

    tmp1 = tmp = (Group *) xmalloc((ngroups+1) * sizeof(Group));
    for (i = 0; i < ngroups; i++, tmp++) {
	    BVAL(tmp) = AVAL(tmp) = (Bit) - 1;
    }
    AVAL(tmp) = MaskRight1(nbits);
    BVAL(tmp) = MaskRight1(nbits);

    return tmp1;
}

/* Memory allocation routines for pass3_decl.  Allocate the memory and
   initialize to all Zs for Nets. */

// ??? mdh - needs fixed should not put ones in all bits of msb
Group *malloc_Z(nbits_t nbits)
{
    ngroups_t ngroups = bits_to_groups(nbits);
    ngroups_t i;
    Group *tmp, *tmp1;

    tmp1 = tmp = (Group *) xmalloc((ngroups+1) * sizeof(Group));
    for (i = 0; i < ngroups; i++, tmp++) {
	    BVAL(tmp) = (Bit) - 1;
	    AVAL(tmp) = 0;
    }
    AVAL(tmp) = 0;
    BVAL(tmp) = MaskRight1(nbits);

    return tmp1;
}

// ??? mdh - needs fixed should not put ones in all bits of msb
void decl_malloc_Z( tree decl )
{
    if( PORT_COLLAPSED_ATTR(decl) ) {
	ASSERT(!STMT_SURROGATE_ATTR(decl));
	return;
    }

    int ngroups = bits_to_groups(TREE_NBITS(decl));
    if(  ngroups >= 1 ) {
        DECL_STORAGE(decl) = (Group *) malloc_Z(TREE_NBITS(decl));
    } else  {
	DECL_STORAGE(decl) = &DECL_DSTORAGE(decl);
	AVAL(DECL_STORAGE(decl)) = 0;
	BVAL(DECL_STORAGE(decl)) = MaskRight1(TREE_NBITS(decl));
    }
}
void decl_malloc_X( tree decl )
{
    if( PORT_COLLAPSED_ATTR(decl) ) {
	ASSERT(!STMT_SURROGATE_ATTR(decl));
	return;
    }
    int ngroups = bits_to_groups(TREE_NBITS(decl));
    if( ngroups >= 1 ) {
        DECL_STORAGE(decl) = (Group *) malloc_X(TREE_NBITS(decl));
    } else {
	DECL_STORAGE(decl) = &DECL_DSTORAGE(decl);
	AVAL(DECL_STORAGE(decl)) = MaskRight1(TREE_NBITS(decl));
	BVAL(DECL_STORAGE(decl)) = MaskRight1(TREE_NBITS(decl));
    }
}

static ngroups_t do_vector_stuff(tree decl)
{
    nbits_t nbits;
    ngroups_t ngroups;

    MSB(decl) =
	get_range(DECL_MSB(decl), IDENTIFIER_POINTER(DECL_NAME(decl)));
    LSB(decl) =
	get_range(DECL_LSB(decl), IDENTIFIER_POINTER(DECL_NAME(decl)));
    nbits = ABS(MSB(decl) - LSB(decl)) + 1;
    VECTOR_DIRECTION_ATTR(decl) = (MSB(decl) >= LSB(decl));
    ngroups = bits_to_groups(nbits);
    if (ngroups > MAX_GROUPS)
	sorry1("Maximum vector size is %d", (char *) MAX_BITS);
    TREE_NBITS(decl) = nbits;
    if( PORT_COLLAPSED_ATTR(decl) ) {
    	nbits_t msb = get_range(DECL_COLLAPSED_MSB(decl), 
			IDENTIFIER_POINTER(DECL_NAME(decl)));
    	nbits_t lsb = get_range(DECL_COLLAPSED_LSB(decl), 
			IDENTIFIER_POINTER(DECL_NAME(decl)));
        nbits_t width = ABS(msb - lsb) + 1;
        DECL_COLLAPSED_NBITS(decl) = width;
    }
    return ngroups;
}

/* Allocate storage for declaration block after evaluating and checking
   LSB and MSBs. */

void pass3_decl(tree decl)
{
    enum tree_type code = (enum tree_type) TREE_CODE(decl);
    nbits_t nbits;
    ngroups_t ngroups;
    ngroups_t array_size;

    lineno = DECL_SOURCE_LINE(decl);
    input_filename = DECL_SOURCE_FILE(decl);
    INITIALIZED_ATTR(decl) = 1;
// ??? mdh - this cast works around a compiler bug in gcc 4.4.1
    switch (static_cast<int>(code)) {
    case NET_SCALAR_DECL:
	pass3_delay(NET_DELAY(decl));

	if (STMT_SURROGATE_ATTR(decl)) {
	    /* copy stuff from original; TREE_CHAIN points to original */
	    TREE_NBITS(decl) = 1;
	    decl_malloc_Z(decl);
	    marker_info.first = NULL;
	    marker_info.last = NULL;
	    marker_info.flags = (enum marker_flags) (M_NET + M_FIXED);
	    if (PORT_IMMEDIATE_ATTR(decl))	
		marker_info.flags =
		    (enum marker_flags) (M_IMMED + marker_info.flags);
	    marker_info.delay = NET_DELAY(decl);
	    if (!NET_ASSIGN_ATTR(decl)
		|| PORT_MIRRORED_ATTR(TREE_CHAIN(decl))
		|| NET_DELAY(TREE_CHAIN(decl)))
		BuildMarker(decl, &marker_info);
	    break;
	}			/* if not surrogate, do the rest the same as for regs */
	TREE_NBITS(decl) = 1;
	decl_malloc_Z(decl);
	break;

    case REG_SCALAR_DECL:
	TREE_NBITS(decl) = 1;
	decl_malloc_X(decl);
	break;

    case INTEGER_DECL:
	MSB(decl) = 31;
	LSB(decl) = 0;
	VECTOR_DIRECTION_ATTR(decl) = 1;
	TREE_NBITS(decl) = 32;
	decl_malloc_X(decl);
	break;

    case REAL_DECL:
	MSB(decl) = 31;
	LSB(decl) = 0;
	VECTOR_DIRECTION_ATTR(decl) = 1;
	TREE_NBITS(decl) = 32;
	decl_malloc_X(decl);
	REAL_(DECL_STORAGE(decl)) = 0;
	break;

    case ARRAY_DECL:
	ARRAY_HI(decl) = get_range(ARRAY_HI_EXPR(decl),
				   IDENT(DECL_NAME(decl)));
	ARRAY_LO(decl) = get_range(ARRAY_LO_EXPR(decl),
				   IDENT(DECL_NAME(decl)));

	switch (ARRAY_CODE(decl)) {
	case REG_VECTOR_DECL:
	    ngroups = do_vector_stuff(decl);
	    break;

	case INTEGER_DECL:
	    TREE_NBITS(decl) = 32;
	    ngroups = 0;
	    break;

	case TIME_DECL:
	    TREE_NBITS(decl) = 64;
	    ngroups = 1;
	    break;

	case REG_SCALAR_DECL:
	    TREE_NBITS(decl) = 1;
	    ngroups = 0;
	    break;

	default:
	    fatal("Internal error: unknown array type", NULL);
	}
	{
	    nbits_t tmp_array_size;
	    tmp_array_size = ABS(ARRAY_HI(decl) - ARRAY_LO(decl)) + 1;
	    if (tmp_array_size > MAX_ARRAY
		|| tmp_array_size * (ngroups + 1) > MAX_ARRAY) {
		error("Maximum array size exceeded in array '%s'",
		      IDENT(DECL_NAME(decl)), NULL_CHAR);
		array_size = 1;
	    } else
		array_size = tmp_array_size;
	}			/* Allocate pointers to array elements */
	DECL_STORAGE(decl) =
	    (Group *) malloc_X(array_size * (ngroups + 1) * 32);

	VECTOR_DIRECTION_ATTR(decl) = (ARRAY_HI(decl) >= ARRAY_LO(decl));
	break;

    case NET_VECTOR_DECL:
	pass3_delay(NET_DELAY(decl));

	if (STMT_SURROGATE_ATTR(decl)) {
	    if (TREE_CODE(TREE_CHAIN(decl)) == NET_SCALAR_DECL)
		fatal("Scalar converted to vector", NULL);
	    /* copy stuff from original; TREE_CHAIN points to original */
	    MSB(decl) = MSB(TREE_CHAIN(decl));
	    LSB(decl) = LSB(TREE_CHAIN(decl));
	    VECTOR_DIRECTION_ATTR(decl) =
		VECTOR_DIRECTION_ATTR(TREE_CHAIN(decl));
	    TREE_NBITS(decl) = TREE_NBITS(TREE_CHAIN(decl));
	    ngroups = bits_to_groups(TREE_NBITS(decl));
	    decl_malloc_Z(decl);
	    marker_info.first = NULL;
	    marker_info.last = NULL;
	    marker_info.flags = (enum marker_flags) (M_NET + M_FIXED);
//  if (/*PORT_INPUT_ATTR (decl) ||*/ PORT_OUTPUT_ATTR (decl))
	    if (PORT_IMMEDIATE_ATTR(decl))	// || NET_ASSIGN_ATTR (decl))
		marker_info.flags =
		    (enum marker_flags) (M_IMMED + marker_info.flags);
	    marker_info.delay = NET_DELAY(decl);
	    if (!NET_ASSIGN_ATTR(decl)
		|| PORT_MIRRORED_ATTR(TREE_CHAIN(decl)))
		BuildMarker(decl, &marker_info);
	    break;
	}
	/* if not surrogate, do the rest the same as for regs */
	ngroups = do_vector_stuff(decl);
	decl_malloc_Z(decl);
	break;

    case TIME_DECL:
	MSB(decl) = 0;
	LSB(decl) = 0;
	TREE_NBITS(decl) = 64;
	DECL_STORAGE(decl) = (Group *) malloc_X(64);
	break;

    case REG_VECTOR_DECL:
	ngroups = do_vector_stuff(decl);
	decl_malloc_X(decl);
	break;

    case TMP_DECL:
	TREE_NBITS(decl) = TREE_NBITS(TREE_CHAIN(decl));	/* inherit bits */
	ngroups = bits_to_groups(TREE_NBITS(decl));
	decl_malloc_Z(decl);
	break;

    case EVENT_DECL:
	TREE_NBITS(decl) = 1;
	decl_malloc_X(decl);
	AVAL(DECL_STORAGE(decl)) = 1;
	BVAL(DECL_STORAGE(decl)) = 0;
	break;
    case BLOCK_DECL:
	TREE_NBITS(decl) = 1;
	break;

    case SPECPARAM_DECL:
    case PARAM_DECL:
	if (!DECL_PARAM_REDIRECT(decl)) {
	    DECL_PARAM_RVAL_CODE(decl) = pass3_expr(DECL_PARAM_RVAL(decl));
	    nbits = TREE_NBITS(DECL_PARAM_RVAL(decl));
	    TREE_REAL_ATTR(decl) = TREE_REAL_ATTR(DECL_PARAM_RVAL(decl));
	    TREE_INTEGER_ATTR(decl) =
		TREE_INTEGER_ATTR(DECL_PARAM_RVAL(decl));
	} else {
	    DECL_PARAM_RVAL_CODE(decl) =
		pass3_expr(DECL_PARAM_REDIRECT(decl));
	    nbits = TREE_NBITS(DECL_PARAM_REDIRECT(decl));
	    TREE_REAL_ATTR(decl) =
		TREE_REAL_ATTR(DECL_PARAM_REDIRECT(decl));
	    TREE_INTEGER_ATTR(decl) =
		TREE_INTEGER_ATTR(DECL_PARAM_REDIRECT(decl));
	}
	ngroups = bits_to_groups(nbits);
	if (ngroups > MAX_GROUPS)
	    sorry1("Maximum vector size is %d", (char *) MAX_BITS);
	if (DECL_MSB(decl)) {
	    do_vector_stuff(decl);
//        MSB (decl) = get_range (DECL_MSB (decl), IDENTIFIER_POINTER (DECL_NAME (decl)));
//        LSB (decl) = get_range (DECL_LSB (decl), IDENTIFIER_POINTER (DECL_NAME (decl)));
//        VECTOR_DIRECTION_ATTR (decl) = (MSB (decl) >= LSB (decl));
	} else {
	    VECTOR_DIRECTION_ATTR(decl) = 1;	/* default to msb>lsb */
	    LSB(decl) = 0;
	    MSB(decl) = nbits - 1;
	}

	TREE_NBITS(decl) = nbits;
	if (bits_to_groups(nbits) > MAX_GROUPS)
	    sorry1("Maximum vector size is %d", (char *) MAX_BITS);
	decl_malloc_X(decl);
	INITIALIZED_ATTR(decl) = 1;
	{
	    /* need to allocate space for expressions; these are defined in
	       pass3.c */
	    extern ngroups_t stack_size;
	    extern int max_label;

	    if (!R_alloc(max_label, stack_size))
		fatal
		    ("Not enough memory to evaluate parameter expression of '%s'",
		     IDENT(DECL_NAME(decl)));
	}
	eval(DECL_PARAM_RVAL_CODE(decl));
	store(decl, decl);
	break;

    default:
	error("Don't know how to allocate for %s (%s)",
	      tree_code_name[code], IDENT(DECL_NAME(decl)));
    }

//    if ((enum tree_code)code != EVENT_DECL && (enum tree_code)code != BLOCK_DECL)
//	if (DECL_STORAGE(decl) == 0)
//	    fatal("Not enough memory to allocate storage for '%s'",
//		  IDENTIFIER_POINTER(DECL_NAME(decl)));

}


/* A module wasn't found in the hierarchy; look for a library using the
   -y option and the +libext options if they are supplied */

tree check_library(char *name)
{
    char *ytmp;
    char *p;
    char *pathend;
    char path[256];
    LibPath_t *ypath;
    File *f;
    tree tmp_module_list, tmp_module;
    tree saved_scope;



    for (ypath = ypathList; ypath != NULL; ypath = ypath->next) {
	strcpy(path, ypath->path);	/* copy the library path */
	strcat(path, PATHSEP);	/* Add trailing "/" (or "\" or ":") */
	strcat(path, name);	/* Add name of module */
	pathend = &path[strlen(path)];

	if (ylibext) {		/* +libext option supplied */
	    ytmp = (char *) xmalloc(strlen(ylibext) + 1);
	    strcpy(ytmp, ylibext);

	    /* loop through all the extensions given in the option */
	    for (p = strtok(ytmp, "+"); p; p = strtok(NULL, "+")) {
		strcpy(pathend, p);	/* append various extensions */
//              printf ("Attempting to open %s\n", path);
		f = File::fopen(path, "rt");	/* try it */
		if (f) {
		    goto foundit;
		}
	    }
	}
    }

    return NULL_TREE;

  foundit:

    simulator.push_stream(fin, 2);
    fin = f;
    char* buffer = xmalloc(strlen(path) + 1);
    strcpy(buffer, path);
    input_filename = buffer;
    lineno = 1;

    tmp_module_list = module_list;	/* Start new module list for... */
    module_list = NULL_TREE;	/* ...library file */

/* We have to be at the top module level for an instance to happen.  Pop
   this scope level so that all idents have no current value.  Now, the
   parse can happen in a clean environment.  Else, the parse will pick up
   idents used in the parent module.  */

#if 0
    saved_scope = current_scope;
    pop_scope();		/* free up idents */
    current_scope = NULL_TREE;	/* incase timescale is used */
    prog_parse();		/* Parse the library file */
    module_list = nreverse(module_list);	/* put first mod first */
    tmp_module = TREE_PURPOSE(module_list);	/* get first one */
    module_list = chainon(tmp_module_list, module_list);	/* reassemble */

    fin = pop_stream();
    set_scope(saved_scope);	/* restore parent scope */
#endif

    parse_at_top_scope(current_scope);

    module_list = nreverse(module_list);	/* put first mod first */
    tmp_module = TREE_PURPOSE(module_list);	/* get first one */
    module_list = chainon(tmp_module_list, module_list);	/* reassemble */

    fin = simulator.pop_stream();
    LIB_MODULE_ATTR(tmp_module) = 1;
    return tmp_module;
}

/* unwind the stack and then parse the library at the top level scope;
   restore the state of the scope stack after */

void parse_at_top_scope(tree scope)
{
    tree new_scope;

    new_scope = pop_scope();
    if (new_scope)
	parse_at_top_scope(new_scope);	/* recursively unwind */

    else {			/* at top, now parse */

	current_scope = NULL_TREE;	/* incase timescale is used */
	tree dummy;
	simulator.prog_parse(dummy);		/* Parse the library file */
    }
    set_scope(scope);		/* unwind recursion */
}



static tree copy_module(tree node)
{
    return copy_block(node);
}

/* Lookup the module being instantiated.  If the module found is a library
   module, compare number of ports to make sure that another library module
   isn't needed.
   Later, this could be converted to use hashing, esp after gates are
   supported */

static tree module_lookup(tree instance)
{
    tree t;
    char *name = IDENT(INSTANCE_MODULE_NAME(instance));

    for (t = module_list; t; t = TREE_CHAIN(t)) {
	if (!strcmp(name, (MODULE_NAME(TREE_PURPOSE(t))))) {
	    if (UDP_ATTR(TREE_PURPOSE(t))) {	/* this is a udp instantiation */
		return (TREE_PURPOSE(t));
	    }


	    if (LIB_MODULE_ATTR(TREE_PURPOSE(t))) {
		if (list_length(INSTANCE_PORTS(instance)) !=
		    list_length(BLOCK_PORTS(TREE_PURPOSE(t))))
		    continue;
	    }
	    return (TREE_PURPOSE(t));
	}
    }
    return check_library(name);	
}

/* Actually connect the ports; emulate collapsing if necessary.  Pass in
   the module port, the instance port argument, and the instance node for
   line number info */

static void do_connect_ports(tree terminal, tree arg_tree, tree instance)
{
    enum tree_code port_code;
    tree arg = TREE_PURPOSE(arg_tree);
    enum tree_code arg_code = TREE_CODE(arg);
    int evalImmediate;
    int collapse;
    tree assign;
    tree port;

    port = TREE_PURPOSE(terminal);
    port_code = TREE_CODE(port);

    if (PORT_REDEFINED_ATTR(port)) {
	port = DECL_THREAD(port);
	port_code = TREE_CODE(port);
    }

    /*
     * Force immediate eval of continious assigns only if both
     * hi and lo connections are nets. Otherwise reg updates 
     * don't get compressed and slow down the simulation
     */

    if ((is_net_code(arg_code) && port_code == arg_code)
	|| (PORT_INPUT_ATTR(port) && PORT_OUTPUT_ATTR(port))) {
	evalImmediate = 1;
    } else {
	evalImmediate = 0;
    }

    if ((is_net_code(arg_code) && is_net_code(port_code))
	|| (PORT_INPUT_ATTR(port) && PORT_OUTPUT_ATTR(port))) {
	// ??? mdh - need to remove this constraint for INOUT's to have
	// a full solution
	collapse = (port_code == arg_code) &&
		   !NO_OPTIMIZE_ATTR(port) && !NO_OPTIMIZE_ATTR(arg);
    } else {
	collapse = 0;
    }

    /*
     * Merge net sources to collapsed net. Redirect collapsee
     * to collapsed net
     */
    if( collapse ) {
	// ??? mdh - this need much more work
	if( !NET_DELAY(arg) && !NET_DELAY(port) &&
            TREE_TYPE(arg) == TREE_TYPE(port) ) {
	    PORT_COLLAPSED_ATTR(port) = 1;
            DECL_COLLAPSED_NET(port) = arg;
	    collapsingNets.push_back(port);
            TREE_3RD(arg_tree) = NULL_TREE;
	    return;
	}
    }

/* If port was copied (multiple module instantiation), then this
   would be non-zero which would screw up output ports. So, make sure
   it is NULL.  */

    TREE_3RD(arg_tree) = NULL_TREE;

    if (PORT_INPUT_ATTR(port))	
    {
	assign = build_cont_assign(check_lval_port(port),
				   arg, DECL_SOURCE_LINE(port), NULL_TREE,
				   0);
	PORT_INPUT_ATTR(STMT_ASSIGN_LVAL(assign)) = PORT_INPUT_ATTR(port);
	PORT_OUTPUT_ATTR(STMT_ASSIGN_LVAL(assign)) =
	    PORT_OUTPUT_ATTR(port);
	PORT_IMMEDIATE_ATTR(STMT_ASSIGN_LVAL(assign)) = evalImmediate;
	/* save port connection in 3rd node og arg's tree_list */
	TREE_3RD(arg_tree) = build_tree_list(assign, NULL_TREE);
    }
    if (PORT_OUTPUT_ATTR(port))	
    {
	if (!is_all_net(arg)) {
	    error("Illegal output port specification: '%s'",
		  IDENT(DECL_NAME(arg)), NULL_CHAR);
	    return;
	} else if ((port_code == NET_VECTOR_DECL
		    || port_code == NET_SCALAR_DECL) &&
		   PORT_INPUT_ATTR(port) ) {
	    tree tmp_decl = copy_node(port);
	    PORT_COLLAPSED_ATTR(tmp_decl) = 0;

	    DECL_THREAD(port) = tmp_decl;
	    DECL_THREAD(tmp_decl) = port;
	    DECL_NAME(tmp_decl) = DECL_NAME(port);

	    assign = build_cont_assign(check_lval_port(arg),
				       tmp_decl, INSTANCE_LINE(instance),
				       NULL_TREE, 0);
	    PORT_MIRRORED_ATTR(port) = 1;
	} else {
	    assign = build_cont_assign(check_lval_port(arg),
				       port, INSTANCE_LINE(instance),
				       NULL_TREE, 0);
	}

	set_immediate_attr(STMT_ASSIGN_LVAL(assign), evalImmediate);
	/* If this is bi-directional, then TREE_LIST will already have been
	   created */
	if (TREE_3RD(arg_tree))
	    TREE_VALUE(TREE_3RD(arg_tree)) = assign;
	else
	    TREE_3RD(arg_tree) = build_tree_list(NULL_TREE, assign);
    }
}

static void connect_ports(tree inside, tree outside)
{
    tree port, arg;

    arg = INSTANCE_PORTS(outside);

    /* named ports match in one of two conditions: if the name matches the
       name of the decl of the port (if it is a decl and not a concat,
       part-select, etc), or if the port is also named and the names match.
       Note that matching is determined if the IDENTIFIER_NODEs are the
       same since there is one hash table for an entire model. */

    if (arg && TREE_VALUE(arg)) {
	for (; arg; arg = TREE_CHAIN(arg)) {
	    for (port = MODULE_PORT_LIST(inside); port;
		 port = TREE_CHAIN(port)) {
		if ((*tree_code_type[TREE_CODE(TREE_PURPOSE(port))] == 'd'
		     && TREE_VALUE(arg) == DECL_NAME(TREE_PURPOSE(port)))
		    || TREE_VALUE(arg) == TREE_VALUE(port)) {
		    if (NAMED_PORT_ATTR(port))
			error("More than one connection to a port",
			      NULL_CHAR, NULL_CHAR);
		    NAMED_PORT_ATTR(port) = 1;
		    do_connect_ports(port, arg, outside);
		    break;
		}
	    }
	    if (!port)
		error("No port named '%s'", IDENT(TREE_VALUE(arg)),
		      NULL_CHAR);
	}

	/* reset named port markers */
	int portNumber = 1;
	for (port = MODULE_PORT_LIST(inside); port;
	     port = TREE_CHAIN(port)) {
	    if( TREE_PURPOSE(port) ) {
	        DECL_PORT_NUMBER(TREE_PURPOSE(port)) = portNumber++;
                DECL_SOURCE_LINE(TREE_PURPOSE(port)) = lineno;
                DECL_SOURCE_FILE(TREE_PURPOSE(port)) = input_filename;
	    }
	    NAMED_PORT_ATTR(port) = 0;
	}
    }

/* Ordered list of ports */
    else {
	int portNumber = 1;
	for (port = MODULE_PORT_LIST(inside); port && arg;
	     port = TREE_CHAIN(port), arg = TREE_CHAIN(arg)) {
	    DECL_PORT_NUMBER(TREE_PURPOSE(port)) = portNumber++;
            DECL_SOURCE_LINE(TREE_PURPOSE(port)) = lineno;
            DECL_SOURCE_FILE(TREE_PURPOSE(port)) = input_filename;
	    if (TREE_PURPOSE(arg))
		do_connect_ports(port, arg, outside);
	}
    }
    if (arg && TREE_PURPOSE(arg))
	warning("Too many instance ports", NULL, NULL);
    else if (port)
	warning("Too few instance ports", NULL, NULL);

}

/* If there are any defparams in the module, they are chained in
   MODULE_DEFPARAMS.  Make sure lval exists, then redirect its parameter
   to our lval.  Only call this for modules. */

static void resolve_defparams(tree scope)
{
    tree t, param, lval;
    tree resolve_hierarchical_name(tree);
    tree search_up_decl(char *, tree);

    for (t = MODULE_DEFPARAMS(scope); t; t = TREE_CHAIN(t)) {
	lval = STMT_ASSIGN_LVAL(t);
	if (HIERARCHICAL_ATTR(lval))
	    param = resolve_hierarchical_name(lval);
	else
	    param = search_up_decl(IDENT(lval), scope);
	if (param == error_mark_node)
	    return;
	if (!param) {
	    error("Parameter '%s' not declared", IDENT(lval), NULL_CHAR);
	    return;
	}
	DECL_PARAM_REDIRECT(param) = STMT_ASSIGN_RVAL(t);
    }
}

/* Initialize all decls in the current scope and all others below.  These
   will be task, functions, and named blocksm and, of course, modules. */

static void initialize_decls(tree scope)
{
    tree t, t1;

    current_scope = scope;

    if (TREE_CODE(scope) == MODULE_BLOCK) {
	resolve_defparams(scope);

	for (t = MODULE_SPECDEFS(scope); t; t = TREE_CHAIN(t)) {
	    if (TREE_CODE(t) == SPECPARAM_DECL && !INITIALIZED_ATTR(t)) {
		pass3_decl(t);
	    }
	}
    }

    for (t = BLOCK_DECL(scope); t; t = TREE_CHAIN(t)) {
	pass3_decl(t);
	if( PORT_COLLAPSED_ATTR(t) ) {
	    continue;
	}
	/* process other sources to a net */
	if (TREE_CODE(t) == NET_VECTOR_DECL ||
	    TREE_CODE(t) == NET_SCALAR_DECL) {
	    marker_info.current_scb = SCB::BuildSCB(t, NOLIST);
	    if (PORT_OUTPUT_ATTR(t) || PORT_MIRRORED_ATTR(t))
		if (DECL_THREAD(t))
		    pass3_decl(DECL_THREAD(t));

	    for (t1 = NET_SOURCE(t); t1; t1 = NET_SOURCE(t1)) {
/*
		if (strcmp(IDENTIFIER_POINTER(DECL_NAME(t1)),
			   IDENTIFIER_POINTER(DECL_NAME(t))))
		    fatal("Line %lu  Driver names to not agree\n",
			  (char *) lineno);
*/
		pass3_decl(t1);
	    }
	}
    }

    for (t = BLOCK_PORTS(scope); t; t = TREE_CHAIN(t)) {
	if(PORT_REDEFINED_ATTR(t)) {
	    continue;
	}
	pass3_decl(t);
	if(PORT_COLLAPSED_ATTR(t)) {
	    continue;
	}
	/* process other sources to a net */
	if (TREE_CODE(t) == NET_VECTOR_DECL ||
	    TREE_CODE(t) == NET_SCALAR_DECL) {
	    marker_info.current_scb = SCB::BuildSCB(t, NOLIST);
	    if (PORT_OUTPUT_ATTR(t) || PORT_MIRRORED_ATTR(t))
		if (DECL_THREAD(t))
		    pass3_decl(DECL_THREAD(t));
    	    if( PORT_COLLAPSED_ATTR(t) ) {
		ASSERT( !NET_SOURCE(t) );
    	        continue;
    	    }

	    for (t1 = NET_SOURCE(t); t1; t1 = NET_SOURCE(t1)) {
		pass3_decl(t1);
	    }
	}
    }

    /* Allocate return variable */
    if (TREE_CODE(scope) == FUNCTION_BLOCK) {
	pass3_decl(FUNCT_DECL(scope));
	TREE_NBITS(scope) = TREE_NBITS(FUNCT_DECL(scope));
    }


    /* do this for everyone below us */
    for (t = BLOCK_DOWN(scope); t; t = TREE_CHAIN(t))
	initialize_decls(t);

}

/* recursively link module together hierarchically */

static void do_instantiation(tree node)
{
    tree t, instance;
    int had_to_copy;

/* Allocate space for declared vars */

#if VDEBUG != 0
    if (vdebug)
	printf_V("****** PASS2: Module '%s', instance '%s'\n",
		 MODULE_NAME(node), IDENT(BLOCK_NAME(node)));
#endif

    INITIALIZED_ATTR(node) = 1;

    /* scan all instantiations within this module */
    for (t = MODULE_INSTANCES(node); t; t = TREE_CHAIN(t)) {
	HEAPTEST("In do_instantiation");
	if (TREE_CODE(t) != INSTANCE_NODE)
	    continue;
	/* get line and file info in case of errors */
	lineno = STMT_SOURCE_LINE(t);
	input_filename = STMT_SOURCE_FILE(t);

	/* first, look for module name */
	instance = module_lookup(t);
	if (!instance) {
	    error("Module '%s' not defined",
		  IDENT(INSTANCE_MODULE_NAME(t)), NULL_CHAR);
	    continue;
	}

	/* if module instance is really a udp, handle here */

	if (UDP_ATTR(instance)) {
	    if (UDP_ATTR(t) == TRUE)	/* already set; must have been copied... */
		continue;	/* ...copy already build the gate instance */
	    UDP_ATTR(t) = TRUE;
	    instantiate_udp(node, t, instance);
	    continue;
	}


	/* Next, if the module is not top-level, the we need to make a copy */
	if (BLOCK_UP(instance)) {
	    instance = copy_module(instance);
	    had_to_copy = 1;
	    if (!instance)
		break;
	} else
	    had_to_copy = 0;

	if (!INSTANCE_NAME(t) && !LIB_MODULE_ATTR(instance)) {
	    error("No module name given in instantiation", NULL_CHAR,
		  NULL_CHAR);
	}
	make_block_decl(INSTANCE_NAME(t), node, instance);

	/* Thread the module in */
	TREE_CHAIN(instance) = BLOCK_DOWN(node);
	BLOCK_DOWN(node) = instance;
	BLOCK_UP(instance) = node;
	BLOCK_NAME(instance) = INSTANCE_NAME(t);

	INSTANCE_BLOCK(t) = instance;

	/* resolve module parameter overrides #(...) */
	if (INSTANCE_PARAMS(t)) {
	    tree decl, param_list = INSTANCE_PARAMS(t);

	    /* scan decl list for a parameter */
	    for (decl = BLOCK_DECL(instance); decl;
		 decl = TREE_CHAIN(decl)) {
		if (TREE_CODE(decl) != PARAM_DECL)
		    continue;
		DECL_PARAM_REDIRECT(decl) = TREE_PURPOSE(param_list);
		if ((param_list = TREE_CHAIN(param_list)) == 0)
		    break;
	    }
	    if (param_list)
		error("Too many module instance parameter assignments",
		      NULL_CHAR, NULL_CHAR);
	}

	connect_ports(instance, t);

	/* Recursively instantiate modules inside the instance (if not already
	   done). */
	if (!INITIALIZED_ATTR(instance)) {
	    set_scope(instance);
	    do_instantiation(instance);
	    current_scope = pop_scope();
	}
    }
}

/* Main entrypoint */

void build_hierarchy()
{
    tree t, t1;
    extern tree scope0;		/* defined in decl.c */

    obstack_init(&inst_obstack);
    obstack_init(&alt_inst_obstack);

    initialize_scope(NULL_TREE);

    for (t1 = module_list; t1; t1 = TREE_CHAIN(t1)) {
	t = TREE_PURPOSE(t1);
	if (!BLOCK_UP(t) && !UDP_ATTR(t)) {	/* don't do anything if not top_level or udp */
	    set_scope(t);
	    do_instantiation(t);
	    current_scope = pop_scope();
	}
    }
    scope0 = make_node(SUPER_BLOCK);
    BLOCK_DECL(scope0) = NULL;

//printf_V ("Allocating variable storage\n");
    for (t1 = module_list; t1; t1 = TREE_CHAIN(t1)) {
	t = TREE_PURPOSE(t1);
	if (!BLOCK_UP(t) && !UDP_ATTR(t)) {	/* don't do anything if not top_level or udp */
	    TREE_CHAIN(t) = top_level;
	    top_level = t;
	    make_block_decl(BLOCK_NAME(t), scope0, t);
	}
    }


    /*
     * fix up collapsed nets:
     * 1) move net sources to collapsed net
     */
   list<tree>::iterator ptr;
   for( ptr = collapsingNets.begin(); ptr != collapsingNets.end(); ++ptr ) {  
	tree collapsing = *ptr;
	tree collapsed = DECL_COLLAPSED_NET(collapsing);
	while(1) {
	    if( !PORT_COLLAPSED_ATTR(collapsed) ) {
		break;
	    }
	    if( collapsed == DECL_COLLAPSED_NET(collapsed) ) {
	        ASSERT( collapsed != DECL_COLLAPSED_NET(collapsed) );
	    }
	    collapsed = DECL_COLLAPSED_NET(collapsed);
	}
	tree src1 = NET_SOURCE(collapsed);
	tree src2 = NET_SOURCE(collapsing);
	for( tree t2 = src2; t2; t2 = NET_SOURCE(t2) ) {
	    TREE_CHAIN(t2) = collapsed;
	}
	if( src1 ) {
	    for( ; NET_SOURCE(src1); src1 = NET_SOURCE(src1) );
	    NET_SOURCE(src1) = src2;
	} else {
	    NET_SOURCE(collapsed) = src2;
	}
	NET_SOURCE(collapsing) = NULL;
	DECL_COLLAPSED_MSB(collapsing) = DECL_MSB(collapsed);
	DECL_COLLAPSED_LSB(collapsing) = DECL_LSB(collapsed);
	if( TREE_CODE(collapsed) == NET_SCALAR_DECL) {
	    DECL_COLLAPSED_NBITS(collapsing) = 1;
	}
	/*
 	 * don't allow collapsed nets to be immediately evaluated.
 	 */
	for( tree t2 = NET_SOURCE(collapsed); t2; t2 = NET_SOURCE(t2) ) {
	    NET_ASSIGN_ATTR(t2) = 0;
	}
   }

    /* Do this after all top-levels are resolved so that hierarchical defparams
       can be accurately located */
    for (t = top_level; t; t = TREE_CHAIN(t))
	initialize_decls(t);

    /*
     * fix up collapsed nets:
     * 1) alias decl storage for collapsing nets to collapsed nets
     */
   for( ptr = collapsingNets.begin(); ptr != collapsingNets.end(); ++ptr ) {  
	tree collapsing = *ptr;
	tree collapsed = DECL_COLLAPSED_NET(collapsing);
	while(1) {
	    if( !PORT_COLLAPSED_ATTR(collapsed) ) {
		break;
	    }
	    collapsed = DECL_COLLAPSED_NET(collapsed);
	}
	ASSERT( DECL_STORAGE(collapsed) );
	DECL_STORAGE(collapsing) = DECL_STORAGE(collapsed);
	if( DECL_COLLAPSED_NBITS(collapsing) != TREE_NBITS(collapsing) ) {
                lineno = DECL_SOURCE_LINE(collapsing);
                input_filename = DECL_SOURCE_FILE(collapsing);
		warning("Port sizes don't match in port #%d",
				(char*)DECL_PORT_NUMBER(collapsing), NULL_CHAR);
	}
   }


    BLOCK_DOWN(scope0) = top_level;
    BLOCK_UP(scope0) = NULL;
    BLOCK_NAME(scope0) = NULL;

}


void print_top_modules()
{
    tree t;

    /* for any module that has not moved, thread it onto the top_level list */
    printf_V("Top-level modules:\n");
    for (t = top_level; t; t = TREE_CHAIN(t))
	printf_V("   %s\n", IDENT(BLOCK_NAME(t)));
    printf_V("\n");
}

/**************************************************************
 *
 *		init_pass2
 *			- initializes global and static variables in pass2.c
 *
 **************************************************************
 */

void init_pass2()
{
    port_connections = NULL_TREE;
}

}
