/*****************************************************************************
 * Copyright (C) 1997-2008, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *	  pli external dll management interface
 *
 ******************************************************************************
 */

#include <string>
#include <map>
#include <list>
#include "pliext.h"
#include "veriuser.h"
#include "io.h"
#include "dirent.h"
#include "ltdl.h"

using namespace veriwell;
using std::string;
using std::map;
using std::list;

/*
 * store all pli routines in a map
 */
map<string,p_tfcell> pliMap;
list<char*>         versionStrings;

/*******************************************************
 *
 * LoadPliLibrary
 * - load pli library dll
 *    returns non-zero on error
 ******************************************************/
int   veriwell::LoadPliLibrary( const char* filename ) 
{ 
    lt_dlhandle handle = lt_dlopen( filename );
    if( !handle ) {
   	return 1;
    }
    /*
     * ignore tool plugins that may be pickedup
     */
    void* be = lt_dlsym( handle, "CreateToolInstance" );
    if( be ) {
        lt_dlclose( handle );
	return 0;
    }

    p_tfcell tfcell = (p_tfcell)lt_dlsym( handle, "veriusertfs" );
    if( !tfcell ) {
        lt_dlclose( handle );
	return 1;
    }
    while( tfcell->type ) {
	pliMap[tfcell->tfname] = tfcell;
	tfcell++;
    } 
    char** vs = (char**)lt_dlsym( handle, "veriuser_version_str" );
    if( vs ) {
	versionStrings.push_back( *vs );
    }
    return 0;
}

/*******************************************************
 *
 * LoadPliPlugins
 * - search directory for pli plugins and load them
 *
 ******************************************************/
void  veriwell::LoadPliPlugins( const char* pluginPath ) 
{ 
    DIR* dir = opendir(pluginPath);
    if( dir ) {
        while(1) {
            struct dirent* entry = readdir( dir );
            if( !entry ) {
                break;
            }
            string path;
            path = pluginPath;
            path += "/";
            path += entry->d_name;
	    LoadPliLibrary( path.c_str() );
        }
    }
}

/*******************************************************
 *
 * DumpPliStrings
 * - dump pli strings to log
 *
 ******************************************************/
void  veriwell::DumpPliStrings()
{
    list<char*>::iterator ptr;
    for( ptr = versionStrings.begin(); ptr != versionStrings.end(); ++ptr ) {
	printf_V( "%s", *ptr );	
    }
}

/*******************************************************
 *
 * FindPliTask
 * - return pli structure for given task
 *
 ******************************************************/
p_tfcell veriwell::FindPliTask( const char* name ) 
{ 
    if( pliMap.find( name ) == pliMap.end() ) {
        return NULL; 
    }
    return pliMap[name];
}




