/******************************************************************************
 * 	Copyright (c) 2008 Mark Hummel DBA Raquette Software.
 *		All rights reserved
 *
 *    This file contains source code written by Raquette Software, 
 *    68 Stewart Street, Franklin MA 02038. It may not be used without 
 *    express written permission. The expression of the information 
 *    contained herein is protected under federal copyright laws and 
 *    all copying without permission is prohibited and may be subject 
 *    to criminal penalties. The Author assumes no responsibility for 
 *    errors, omissions, or damages caused by the use of these programs 
 *    or from use of the information contained herein.
 *
 *-----------------------------------------------------------------------------
 */
/******************************************************************************
 *
 *
 *	   stats.cc
 *		- methods for statistics collection
 *
 ******************************************************************************
 */
						

#include "vtypes.h"
#include "flags.h"
#include "io.h"
#include "stats.h"
using veriwell::printf_V;

MasterStats masterStats;


/**************************************************
 * MasterStats
 * - overall statistics manager
 *************************************************/

static void DumpMasterStats()
{
    masterStats.Dump();
}
/*****************************************
  Constructor
 ****************************************/
MasterStats::MasterStats() {
}
/*****************************************
  Destructor
 ****************************************/
MasterStats::~MasterStats() {
    Dump();
}

/*****************************************
  Add
  - Add statistics widget to master
 ****************************************/
void MasterStats::Add( Stats* widget )
{
    widgets.push_back( widget );
    list<Stats*>::iterator ptr;
}

/*****************************************
  Start
  - Start gathering stats
 ****************************************/
void MasterStats::Start()
{
   startTime = veriwell::rdtsc();
}
/*****************************************
  Dump
  - Dump all stats
 ****************************************/
void MasterStats::Dump()
{
    if( !veriwell::printStats ) {
	return;
    }
    unsigned long long int delta = veriwell::rdtsc() - startTime;
    list<Stats*>::iterator ptr;
    for( ptr = widgets.begin(); ptr != widgets.end(); ++ptr ) {
	(*ptr)->Dump( delta );
    }
    printf_V( "\nOverall Statistics\n" );
    printf_V( "%-20s: %20s %4s\n", "Subsystem", "Total Cycles", "%" );
    printf_V( "-------------------------------------------------\n" );
    unsigned long long totalCycles = 0;
    for( ptr = widgets.begin(); ptr != widgets.end(); ++ptr ) {
	unsigned long long cycles = (*ptr)->TotalCycles();
	const char* name = (*ptr)->Name();
	totalCycles += cycles;
	if( cycles ) {
            printf_V( "%-20s: %20lld %4.2f\n", name, cycles, (double)cycles*100./delta );
        }
    }
    printf_V( "%-20s: %20lld %4.2f\n", "other", delta - totalCycles, (delta-totalCycles)*100./delta );
    printf_V( "%-20s: %20lld\n", "total cycles", delta );
}
