/*****************************************************************************
 * Copyright (C) 1997-2009, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   array.h
 *		- array info routines
 *
 ******************************************************************************
 */

#ifndef ARRAY_H
#define ARRAY_H

#include "cnode_def.h"
#include "cdecl.h"
#include "cgenvar.h"

/**
 * Class method to query array info.
 */
class Array {
public:
	/**
 	 * Get width of array.
 	 * \param ref reference to a decl
 	 * \param index index list
 	 * \return width of expression
 	 */
	static int Width( CNode* ref, CNode* index );
	/**
 	 * Get width of array as an expression.
 	 * \param ref reference to a decl
 	 * \param index index list
 	 * \return width expression.
 	 */
	static CNode* WidthExp( CNode* ref, CNode* index );
	/**
 	 * Determine if width of array variable is constant.
 	 * \todo current not fully implemented.
 	 * \param ref reference to a decl
 	 * \param index index list
 	 * \return non-zero if width is constant. ie dependent upon only
 	 *         on parameters and constants.
 	 */
	static int WidthConstant( CNode* ref, CNode* index );
	/**
 	 * Determine if width of array variable is volatile.
 	 * \todo current not fully implemented.
 	 * \param ref reference to a decl
 	 * \param index index list
 	 * \return non-zero if width is volatile. ie dependent upon only
 	 *         on parameters or variables.
 	 */
	static int WidthVolatile( CNode* ref, CNode* index );
	/**
 	 * Determine if width of array can be evaluated.
 	 * \todo current not fully implemented.
 	 * \param ref reference to a decl
 	 * \param index index list
 	 * \return non-zero if width can be evaluated.
 	 */
	static int WidthEvaluateable( CNode* ref, CNode* index );
	/**
 	 * Determine type of array.
 	 * \todo current not fully implemented.
 	 * \param ref reference to a decl
 	 * \param index index list
 	 * \return type of array return value.
 	 */
	static NodeType_t Type( CNode* ref, CNode* index );
	/**
 	 * Get decl for array reference
 	 * \param ref decl reference node
 	 * \return decl associated with the reference
 	 */
	static CDecl* ArrayDecl( CNode* ref );
	/**
 	 * Get datatype info for array reference
 	 * \param ref reference to a decl/subreference
 	 * \param index index list 
 	 * \param dataType resolved base data type
 	 * \param unpackedDimensions list of unresolved unpacked dimensions
 	 * \param packedDimensions list of unresolved packed dimensions 
 	 * \param residualPackedIndices list of unresolved packed indices
 	 * \return -1 if reference is valid but unresolvable
 	 * 	    0 if reference is invalid
 	 * 	    1 if reference is valid and was resolved
 	 */
	static int ResolveReference( 
		CNode*          ref, 
		CNode*          index,
		CDataType*&     dataType,
		vector<CNode*>& unpackedDimensions,
		vector<CNode*>& packedDimensions,
		vector<CNode*>& residualPackedIndices );
};

#endif // ARRAY_H
