/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cattr.cpp
 *		- class for attribute
 *		  definition nodes
 *
 ******************************************************************************
 */


#include <stdio.h>
#include "glue.h"
#include "cattr.h"
#include "cnode.h"


/**************************************************
  Constructor
 *************************************************/

CAttr::CAttr( Coord_t* aLoc ) 
    : CDecl( NULL, aLoc, eATTR, NULL, eFLAG_NONE )
{}

/**************************************************
  Clone
  - Create a deep copy of attribute
 *************************************************/
CDecl* CAttr::Clone( CObstack* heap )
{
    CAttr* clone = new(heap) CAttr( GetCoord() );
    clone->Copy( heap, *this );
    return clone;
}

/**************************************************
  Copy
  - Deep copy from given attribute
 *************************************************/
void CAttr::Copy( CObstack* heap, CAttr& d )
{
    table.clear();
    vector<Entry>::iterator ptr;
    for( ptr = d.table.begin(); ptr != d.table.end(); ++ptr ) {
	Entry entry;
	entry.s = ptr->s;
	entry.exp = ptr->exp ? ptr->exp->Clone(heap) : NULL;
	table.push_back( entry );
    }
    
}

/*************************************************
  HasAttribute
    - returns 1 if attribute is specified
 ************************************************/
int CAttr::HasAttribute( const char* name )
{
    vector<Entry>::iterator ptr;
    for( ptr = table.begin(); ptr != table.end(); ++ptr ) {
	if( ptr->s == name ) {
	    return 1;
 	}
    }
    return 0;
}

/*************************************************
  GetAttributeExpression
    - returns expression associated with attribute
 ************************************************/
CNode* CAttr::GetAttributeExpression( const char* name )
{
    vector<Entry>::iterator ptr;
    for( ptr = table.begin(); ptr != table.end(); ++ptr ) {
	if( ptr->s == name ) {
	    return ptr->exp;
 	}
    }
    return NULL;
}

/*************************************************
  Dump
    - dump attribute info
 ************************************************/
void CAttr::Dump( FILE* f )
{
    fprintf( f, "ATTRIBUTE:\n" );
}
