/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cenum.h
 *		- class definition of enum constants
 *
 ******************************************************************************
 */

#ifndef CENUM_H
#define CENUM_H

#include <stdio.h>
#include "glue.h"
#include "cnode.h"
#include "cdecl.h"


class CNode;


/**
 * Declaration object for enum constants.
 */
class CEnum: public CDecl
{
    CNode* expression;
public:
	/**
 	 * Create a enum const declaration.
 	 * \param symbol name of declaration.
 	 * \param aLoc file coordinates of declaration.
 	 */
	CEnum( CSymbol* symbol, Coord_t* aLoc, 
	       CDataType* dataType, CNode* expression ) 
	 : CDecl( symbol, aLoc, eENUMVALUE, dataType, eFLAG_NONE ),
	   expression(expression) {}
	/**
 	 * Create a clone of this declaration.
 	 * \param heap heap to use for allocation.
 	 * \return new declaration.
 	 */
	virtual CDecl* Clone( CObstack* heap );
	/**
 	 * Return expression for enum value
 	 * \return enum value as an expresion.
 	 */
	virtual CNode* GetExpression( void ) { return expression; }
	/**
 	 * Dump enum info to file descriptor.
 	 * \param f file descriptor.
 	 */
        virtual void Dump( FILE* f );
protected:
	/**
 	 * Deep copy of declaration.
 	 * \param heap heap to use for allocation.
 	 * \param d declaration to copy.
 	 */
	void Copy( CObstack* heap, CEnum& d );
private:
	/*
 	 * disable copy constructor
 	 */
	CEnum( const CEnum& );
};

#endif // CENUM_H
