/*****************************************************************************
 * Copyright (C) 1997-2010, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cenumtype.cc
 *		- abstract class for enum types 
 *
 *
 ******************************************************************************
 */

#include "cenumtype.h"


/**
 * Create a new copy with a deep copy. 
 * \param heap heap to use for allocation.
 */
CDataType* CEnumType::Clone( CObstack* heap )
{
    list<CNode*> dummy;
    CEnumType* clone = new(heap) CEnumType(GetVarDataType(), NULL, NULL, dummy);
    clone->Copy( heap, *this ); 
    return clone;
}

/**
 * Perform deep copy of given object to this one
 * This should never be call directly, only by subclasses.
 * \param o declaration to copy.
 */
void CEnumType::Copy( CObstack* heap, const CEnumType& o )
{
    CDataType::Copy( heap, o );
    baseType = o.baseType->Clone(heap);
    enumList = o.enumList->Clone(heap);
}
/**
 * Dump data type info to file descriptor.
 * \param f file descriptor.
 */
void CEnumType::Dump( FILE* f ) const
{
    fprintf( f, "enum " );
    if( baseType ) {
	baseType->Dump( f );
    } else {
	fprintf( f, " " );
    }
}
/**
 * Walk tree invoking callback on each node before children have 
 * been visited. callback can terminate transversal early.
 * callback returns 0 to terminate transversal early.
 * \param callback function to be called for each node
 * \param data opaque argument passed to callback function
 */
void CEnumType::PreVisit1( int (*callback)(CNode*,void*), void* data )
{
    CDataType::PreVisit1( callback, data );
    if( baseType ) {
	baseType->PreVisit1( callback, data );
    }
    if( enumList ) {
	enumList->PreVisit1( callback, data );
    }
}
/**
 * Walk tree invoking callback on each node after children have 
 * been visited.
 * \param callback function to be called for each node
 * \param data opaque argument passed to callback function
 */
void CEnumType::PostVisit1( void (*callback)(CNode*,void*), void* data )
{
    CDataType::PostVisit1( callback, data );
    if( baseType ) {
	baseType->PostVisit1( callback, data );
    }
    if( enumList ) {
	enumList->PostVisit1( callback, data );
    }
}
/**
 * Walk tree invoking callback on each node after children have 
 * been visited. callback may modify tree and return a modified
 * subtree.
 * \param callback function to be called for each node
 * \param data opaque argument passed to callback function
 * \return modified child subtree.
 */
void CEnumType::PostSubVisit1( CNode* (*callback)(CNode*, void*), void* data )
{
    CDataType::PostSubVisit1( callback, data );
    if( baseType ) {
	baseType->PostSubVisit1( callback, data );
    }
    if( enumList ) {
	enumList->PostSubVisit1( callback, data );
    }
}

