/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cgate.cpp
 *		- class for managing instances of gates
 *
 *
 ******************************************************************************
 */


#include "cgate.h"
#include "csymbol.h"
#include "cobstack.h"
#include "cmodule.h"
#include "csymtab.h"




/*******************************************
	Constructor
 *******************************************/
	CGate::CGate( CSymbol* name, 
		CSymbol* type, 
		Strength_t s0,
		Strength_t s1,
		CNode* delayControl,
		CNode* arguments,
		Coord_t* loc ) : 
		CDecl( name, loc, eINSTANCE, NULL, eFLAG_ARRAY ),
		type( type ),
		delayControl( delayControl ),
		arguments( arguments ),
		strength0(s0),
		strength1(s1)
{
}

/****************************************************
	Clone
	- return a deep copy of declaration
*****************************************************/
CDecl* CGate::Clone( CObstack* heap )
{
     CGate* clone = new(heap) CGate( GetSymbol(), NULL, eUNDEFINED, eUNDEFINED, 
					NULL, NULL, GetCoord() );
     clone->Copy( heap, *this );
     return clone;
}

/****************************************************
	Copy
	- Deep copy of declaration
*****************************************************/
void CGate::Copy( CObstack* heap, CGate& d )
{
    type         = d.type;
    delayControl = d.delayControl ? d.delayControl->Clone(heap) : NULL;
    arguments    = d.arguments ? d.arguments->Clone(heap) : NULL;
    strength0	 = d.strength0;
    strength1	 = d.strength1;
}
/*******************************************
	GetType	
	- return gate type 
 *******************************************/

CSymbol* CGate::GetType( void )
{
	return type;
}
/*******************************************
	GetDelayControl	
	- return delay control 
 *******************************************/

CNode* CGate::GetDelayControl( void )
{
	return delayControl;
}
/*******************************************
	GetArguments	
	- return arguments 
 *******************************************/

CNode* CGate::GetArguments( void )
{
	return arguments;
}

/****************************************************
    PreVisit1
	- helper routine to support PreVisit1 node 
	  traversal.
*****************************************************/
void CGate::PreVisit1( int (*func)(CNode*,void*), void* data )
{
    if( delayControl ) {
	delayControl->PreVisit1( func, data );
    }
    if( arguments ) {
	arguments->PreVisit1( func, data );
    }

    CDecl::PreVisit1( func, data );
}

/****************************************************
    PostVisit1
	- helper routine to support PostVisit1 node 
	  traversal.
*****************************************************/
void CGate::PostVisit1( void (*func)(CNode*,void*), void* data )
{
    if( delayControl ) {
	delayControl->PostVisit1( func, data );
    }
    if( arguments ) {
	arguments->PostVisit1( func, data );
    }

    CDecl::PostVisit1( func, data );
}

/****************************************************
    PostSubVisit1
	- helper routine to support PostSubVisit1 node 
	  traversal.
*****************************************************/
void CGate::PostSubVisit1( CNode* (*func)(CNode*, void*), void* data )
{
    if( delayControl ) {
	delayControl = delayControl->PostSubVisit1( func, data );
    }
    if( arguments ) {
	arguments = arguments->PostSubVisit1( func, data );
    }

    CDecl::PostSubVisit1( func, data );
}

/*******************************************
	Dump	
	- print out instance info 
 *******************************************/
void	CGate::Dump( FILE* f )
{
	
	fprintf( f, "GATE<%s>: ", type->GetName() );
	CDecl::Dump( f );
	fprintf( f, "\n" );
}


