/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cmacro.cpp
 *		- methods for macro 
 *
 *
 ******************************************************************************
 */


#include <string.h>
#include "cmacro.h"
#include "csymbol.h"
#include "csymtab.h"

extern CSymtab<CMacro>	macroSymtab;



/*******************************************
 	LookupMacro
 *******************************************/

CMacro* CMacro::LookupMacro( const char* aName )
{
	CSymbol*	symbol;
	CMacro*		macro;

	symbol = CSymbol::Lookup( aName );
	macro = macroSymtab.Lookup( symbol );
	return macro;
}

/*******************************************
	Constructor
 *******************************************/
	CMacro::CMacro( CSymbol* symbol, Coord_t* aLoc ) : 
		CDecl( symbol, aLoc, eMACRO, NULL, eFLAG_NONE )
{

	body    = NULL;
	argList = NULL;
	defined = TRUE;
	zone    = 0;
	macroSymtab.Add( symbol, this ); 
}

/****************************************************
	Clone
	- return a deep copy of declaration
*****************************************************/
CDecl* CMacro::Clone( CObstack* heap )
{
     CMacro* clone = new(heap) CMacro( GetSymbol(), GetCoord() );
     clone->Copy( heap, *this );
     return clone;
}

/****************************************************
	Copy
	- Deep copy of declaration
*****************************************************/
void CMacro::Copy( CObstack* heap, CMacro& d )
{
    body    = d.body;
    argList = d.argList;
    complex = d.complex;
    defined = d.defined;
    zone    = d.zone;
}

/*******************************************
	GetBody
	- return macro body	
 *******************************************/
char*	CMacro::GetBody( void )
{
	return body;
}
/*******************************************
	SetBody
	- set macro body	
 *******************************************/
void	CMacro::SetBody( char* aBody )
{
	body = aBody;
}
/*******************************************
	SetComplex
	- mark macro as complex 
 *******************************************/
void	CMacro::SetComplex( int aComplex )
{
	complex = aComplex;
}
/*******************************************
	IsComplex	
	- returns TRUE if macro is complex 
 *******************************************/
int	CMacro::IsComplex(  )
{
	return complex;
}
/*******************************************
	SetDefined	
	- mark macro as Defined 
 *******************************************/
void	CMacro::SetDefined( int aDefined )
{
	defined = aDefined;
}
/*******************************************
	IsDefined	
	- returns TRUE if macro is Defined 
 *******************************************/
int	CMacro::IsDefined( void )
{
	return defined;
}
/*******************************************
	GetArgList
	- return macro argument list 
 *******************************************/
char**	CMacro::GetArgList( void )
{
	return argList;
}
/*******************************************
	SetArgList
	- set argument list 
 *******************************************/
void	CMacro::SetArgList( char** aArgList )
{
	argList = aArgList;
}
/*******************************************
	Dump	
	- print out macro info 
 *******************************************/
void	CMacro::Dump( FILE* f )
{
	int	i;
	
	if( complex ) {
		fprintf( f, "\tCOMPLEX MACRO: " );
	} else {
		fprintf( f, "\tSIMPLE MACRO: " );
	}
	CDecl::Dump( f );
	fprintf( f, "\n\tBody -> %s\n", body );
	if( complex ) {
		for( i = 0; argList[i] ; i++ ) {
			fprintf( f, "\tArg %d -> %s\n", i, argList[i] );
		}
	}
}


