/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cnet.cpp
 *		- class definition of net 
 *		  definition nodes
 *
 ******************************************************************************
 */

#include "cnet.h"
#include "cnode.h"
#include "cvector.h"




/****************************************************
	Constructor
*****************************************************/
CNet::CNet( CSymbol* name, Coord_t* aLoc, Wire_t wireType, 
	    CDataType* dataType, int undefined ) :
	CDecl( name, aLoc, eNET, dataType, Or(eFLAG_ARRAY,eFLAG_DATATYPE) ), 
	undefined(undefined) {
	delay = NULL;
	scalared = 0;
	vectored = 0;
	portdir = NULL;
	strength.s0 = eUNDEFINED;
	strength.s1 = eUNDEFINED;
	SetWireAttr( wireType );
}
/****************************************************
	Clone 
	- create a deep copy of net
*****************************************************/
CDecl* CNet::Clone( CObstack* heap )
{
    CNet* clone = new(heap) CNet( GetSymbol(), GetCoord(),
			GetWireAttr(), NULL, GetUndefined() );
    clone->Copy( heap, *this );
    return clone;
}

/****************************************************
	Copy 
	- perform deep copy
*****************************************************/
void CNet::Copy( CObstack* heap, const CNet& net )
{
	CDecl::Copy( heap, net );
	undefined = net.undefined;
	delay     = net.delay ? net.delay->Clone(heap) : NULL;
	scalared  = net.scalared;
	vectored  = net.vectored;
	strength  = net.strength;
	portdir   = net.portdir;
}

/****************************************************
    PreVisit1
	- helper routine to support PreVisit1 node 
	  traversal.
*****************************************************/
void CNet::PreVisit1( int (*func)(CNode*,void*), void* data )
{
    if( GetDelay() ) {
	GetDelay()->PreVisit1( func, data );
    }

    CDecl::PreVisit1( func, data );
}

/****************************************************
    PostVisit1
	- helper routine to support PostVisit1 node 
	  traversal.
*****************************************************/
void CNet::PostVisit1( void (*func)(CNode*,void*), void* data )
{
    if( GetDelay() ) {
	GetDelay()->PostVisit1( func, data );
    }
    CDecl::PostVisit1( func, data );
}

/****************************************************
    PostSubVisit1
	- helper routine to support PostSubVisit1 node 
	  traversal.
*****************************************************/
void CNet::PostSubVisit1( CNode* (*func)(CNode*, void*), void* data )
{
    if( GetDelay() ) {
	SetDelay( GetDelay()->PostSubVisit1( func, data ) );
    }
    CDecl::PostSubVisit1( func, data );
}

/****************************************************
	Dump
*****************************************************/
void	CNet::Dump( FILE* f )
{
    fprintf( f, "%s ", declName[GetType()] );
    CDataType* dataType = GetDataType();
    if( dataType ) {
	dataType->Dump( f );
    }
    fprintf( f, ": %s, defined in ", GetName() ); 
    CDecl::Dump( f );
    fprintf( f, "\n" );
}	

