/*-----------------------------------------------------------------------------
 * 	Copyright (c) 1997-2004 Mark Hummel DBA Raquette Software.
 *		All rights reserved
 *
 *    This file contains source code written by Raquette Software, 
 *    68 Stewart Street, Franklin MA 02038. It may not be used without 
 *    express written permission. The expression of the information 
 *    contained herein is protected under federal copyright laws and 
 *    all copying without permission is prohibited and may be subject 
 *    to criminal penalties. The Author assumes no responsibility for 
 *    errors, omissions, or damages caused by the use of these programs 
 *    or from use of the information contained herein.
 *
 *-----------------------------------------------------------------------------
 */
/******************************************************************************
 *
 *	   compile.hpp
 *
 ******************************************************************************
 */
 
/**
 * Library of compile related routines
 */

#ifndef COMPILE_PP
#define COMPILE_HPP

#include <set>
#include <map>
using namespace std;

class CNode;
class CModule;
class CDataType;

/**
 * Analyse module declarations in parse tree.
 * \ingroup node_utilities
 * \param codeList parse tree to anaylse.
 * \param allModules add all module declarations to set.
 * \param topLevelModules add all toplevel module declarations to set.
 * \param leafModules add all leaf module declarations to set.
 * \param undefinedModules add undefined module declarations to set.
 * \param module2Comments map to load associating all comments before module
 *                        with the given module.
 */
void AnalyseModules( 
    CNode*                codeList, 
    set<CModule*>&        allModules,
    set<CModule*>&        topLevelModules,
    set<CModule*>&        leafModules,
    set<CModule*>&        undefinedModules,
    map<CModule*,CNode*>& module2Comments
);

/**
 * Get datatype info for array, structure and variable references
 * \param ref reference to a decl/array/member node
 * \param dataType resolved base data type
 * \param unpackedDimensions list of unresolved unpacked dimensions
 * \param packedDimensions list of unresolved packed dimensions 
 * \param residualPackedIndices list of unresolved packed indices
 * \return -1 if reference is valid but unresolvable
 * 	    0 if reference is invalid
 * 	    1 if reference is valid and was resolved
 */
int ResolveReference( 
	CNode*          ref, 
	CDataType*&     dataType,
	vector<CNode*>& unpackedDimensions,
	vector<CNode*>& packedDimensions,
	vector<CNode*>& residualPackedIndices );

#endif // COMPILE_HPP
