/*****************************************************************************
 * Copyright (C) 1997-2013, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cpackage.cc
 *		- class for packages
 *
 ******************************************************************************
 */


#include <stdio.h>
#include "glue.h"
#include "cpackage.h"
#include "cnode.h"

extern CSymtab<CDecl>	cuSymtab;


/**************************************************
  Constructor
 *************************************************/

CPackage::CPackage( CSymbol* symbol, Coord_t* aLoc )
    : CDecl( symbol, aLoc, ePACKAGE, NULL, eFLAG_NONE ),
      codeList( NULL ),
      trailingLabel( CSymbol::Lookup("") )
{
}

/*******************************************
	PreVisit
	- Support routine for PreVisit 
 	  tree transveral routine
 *******************************************/

void CPackage::PreVisit1( int (*func)(CNode*,void*), void* data )
{
    CNode* codeList = GetCodeList();
    if( codeList != NULL ) {
    	codeList->PreVisit1( func, data );	
    }
}

/*******************************************
	PostVisit1
	- Support routine for PostVisit 
 	  tree transveral routine
 *******************************************/

void CPackage::PostVisit1( void (*func)(CNode*,void*), void* data )
{
    CNode* codeList = GetCodeList();
    if( codeList != NULL ) {
    	codeList->PostVisit1( func, data );	
    }
}

/*******************************************
	PostVisit1
	- Support routine for PostVisit 
 	  tree transveral routine
 *******************************************/

void CPackage::PostSubVisit1( CNode* (*func)(CNode*, void*), void* data )
{
    CNode* codeList = GetCodeList();
    if( codeList != NULL ) {
    	SetCodeList( codeList->PostSubVisit1( func, data ) );	
    }
}

/*******************************************
	Resolve
	  Get decl for given scoped reference
	  returns NULL if not found
 *******************************************/
CDecl* CPackage::Resolve( CSymbol* sym )
{
	CDecl* result = NULL;
	const char* s = sym->GetName();
	const char* b1 = strstr( s, "::" );
	if( b1 ) {
	    string pname(s, b1-s );
	    string sname(b1+2);
	    CSymbol* ssym = CSymbol::Lookup( sname.c_str() ); 
	    /*
	     * handle $unit as current scope
	     */
	    if( pname == "$unit" ) {
	        result = cuSymtab.Lookup( ssym );
	    } else {
	        CSymbol* psym = CSymbol::Lookup( pname.c_str() );
	        CDecl* decl = cuSymtab.Lookup( psym );
	        if( decl && decl->GetType() == ePACKAGE ) {
	            CPackage* package = dynamic_cast<CPackage*>(decl);
	            result = package->GetSymtab().Lookup( ssym );
	        }
	    }	
	}
	return result;
}

/*************************************************
  Dump
    - dump package info
 ************************************************/
void CPackage::Dump( FILE* f )
{
    fprintf( f, "PACKAGE: " );
    fprintf( f, "%s, defined in  ", GetName() );
    CDecl::Dump( f );
    fprintf( f, "\n" );
    fprintf( f, "symbol table:\n" );
    GetSymtab().Dump(f,0);
}
