/*****************************************************************************
 * Copyright (C) 1997-2013, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cpackage.h
 *		- class definition of packages
 *
 ******************************************************************************
 */

#ifndef CPACKAGE_H
#define CPACKAGE_H

#include <stdio.h>
#include "glue.h"
#include "cdecl.h"
#include "csymtab.h"

using namespace std;


class CNode;


/**
 * Declaration object for managing packages
 */
class CPackage: public CDecl
{
private:
    	CSymbol* 	symbol;
	CNode*		codeList;
	CSymtab<CDecl>  symtab; 
        CSymbol*	trailingLabel;
public:
	/**
 	 * Create a attribute object
 	 */
	CPackage( CSymbol* symbol, Coord_t* aLoc );
	/**
 	 * Create a clone of this declaration.
 	 * \param heap heap to use for allocation.
 	 * \return new declaration.
 	 */
	virtual CDecl* Clone( CObstack* heap ) { MASSERT( FALSE ); }
	/**
 	 * Get packagess parse tree.
 	 * \return parse tree.
 	 */
	virtual CNode*	GetCodeList() { return codeList; }
	/**
 	 * Set block's parse tree.
 	 * \param list block's parse tree.
 	 */
	virtual void	SetCodeList( CNode* list ) { codeList = list; }
	/**
 	 * Dump attributes to file.
 	 * \param f file descriptor.
 	 */
	virtual void Dump( FILE* f );
	/**
 	 * Set block's symbol table.
 	 * \param symtab reference to new symbol table.
 	 */
	virtual void 	SetSymtab( CSymtab<CDecl>& symtab ) { this->symtab = symtab; }
	/**
 	 * Get block's symbol table.
 	 * \return reference to block's symbol table.
 	 */
	virtual CSymtab<CDecl>& GetSymtab() { return symtab; }
	/**
 	 * Set this block's end label 
 	 * \param s trailing label
 	 */
	void	SetTrailingLabel( CSymbol* s ) { trailingLabel = s; }
	/**
 	 * Get this block's end label
 	 * \return blocks trailing label
 	 */
	CSymbol* GetTrailingLabel( void ) { return trailingLabel; }
	/**
	 * Get decl for give scoped reference
	 * \param sym symbol to be resolved
	 * \return resolved declaration
	 */
	static CDecl* Resolve( CSymbol* sym );

	void    PreVisit1( int (*func)(CNode*,void*), void* data );
	void    PostVisit1( void (*func)(CNode*, void*), void* data );
	void    PostSubVisit1( CNode* (*func)(CNode*, void*), void* data );
private:
	/*
 	 * Disable copy constructor.
 	 */
	CPackage( const CPackage& );
};

#endif // CPACKAGE_HPP
