/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cport.cpp
 *		- class definition of port 
 *		  definition nodes
 *
 ******************************************************************************
 */

#include "cport.h"
#include "cnode.h"




/****************************************************
	Constructor
*****************************************************/
CPort::CPort( CSymbol* name, int dotted, Coord_t* aLoc ) : 
	CDecl( name, aLoc, ePORT, NULL, eFLAG_PRAGMA ),
	dotted(dotted) {
	direction  = (Decl_t)-1;
	connection = NULL;
	wildcard   = false;
}

/****************************************************
	Copy
	- Deep copy from one port to another.
*****************************************************/
void CPort::Copy( CObstack* heap, const CPort& port )
{
    CDecl::Copy( heap, *this );
    connection = port.connection ? port.connection->Clone(heap) : NULL;
    dotted     = port.dotted;
    direction  = port.direction;
    wildcard   = port.wildcard;
}
/****************************************************
	Clone
	- Create a new port that is a deep copy of port
*****************************************************/
CDecl* CPort::Clone( CObstack* heap ) 
{
    CPort* clone = new(heap) CPort( GetSymbol(), IsDotted(), GetCoord() );
    clone->Copy( heap, *this);
    return clone;
}

/****************************************************
	GetConnection
*****************************************************/

CNode*	CPort::GetConnection()

{
	return connection;
}
/****************************************************
	SetConnection
*****************************************************/

void	CPort::SetConnection( CNode* aConnection )
{
	connection = aConnection;
}
/*******************************************
	PreVisit
	- Support routine for PreVisit 
 	  tree transveral routine
 *******************************************/

void CPort::PreVisit1( int (*func)(CNode*,void*), void* data )
{
    if( connection ) {
    	connection->PreVisit1( func, data );	
    }
    CDecl::PreVisit1( func, data );
}
/*******************************************
	PostVisit
	- Support routine for PostVisit 
 	  tree transveral routine
 *******************************************/

void CPort::PostVisit1( void (*func)(CNode*,void*), void* data )
{
    if( connection ) {
    	connection->PostVisit1( func, data );	
    }
    CDecl::PostVisit1( func, data );
}
/*******************************************
	PostSubVisit
	- Support routine for PostSubVisit 
 	  tree transveral routine
 *******************************************/

void CPort::PostSubVisit1( CNode* (*func)(CNode*, void*), void* data )
{
    if( connection ) {
    	connection = connection->PostSubVisit1( func, data );	
    }
    CDecl::PostSubVisit1( func, data );
}

/****************************************************
	Dump
*****************************************************/
void	CPort::Dump( FILE* f )
{
	if( connection == NULL ) {
		fprintf( f, "%s: %s, defined in ", 
			declName[GetType()], GetName() );
	} else {
		fprintf( f, "%s: %s -> ", 
			declName[GetType()], GetName() );
		connection->Dump( f );
		fprintf( f, " defined in " );
	}
	CDecl::Dump( f );
	fprintf( f, "\n" );
}	

