/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cport.hpp
 *		- class definition of port
 *		  definition nodes
 *
 ******************************************************************************
 */

#ifndef CPORT_HPP
#define CPORT_HPP

#include <stdio.h>
#include "glue.h"
#include "cdecl.h"


class CNode;

/**
 * Declaration object for module/function/task ports.
 */
class CPort : public CDecl
{
private:
	CNode*	connection;	///< expression connecting to port.
	int	dotted;		///< non-zero for dotted style ports.
	Decl_t	direction;	///< port direction.
	bool    wildcard;	///< true if port is a wild card
public:
	/**
 	 * Create port.
 	 * \param symbol name of port.
 	 * \param dotted non-zero for dotted style ports.
 	 * \param aLoc file coordinate of port.
 	 */
	CPort( CSymbol* symbol, int dotted, Coord_t* aLoc );
	/**
 	 * Create a clone of this declaration.
 	 * \param heap heap to use for allocation.
 	 * \return new declaration.
 	 */
	virtual CDecl* Clone( CObstack* heap );
	/**
 	 * Dump port info to file descriptor.
 	 * \param f file descriptor.
 	 */
	virtual void Dump( FILE* f );
	/**
 	 * Get connection expression for port.
 	 * \return connection expression.
 	 */
	CNode* GetConnection( void );
	/**
 	 * Set connection expression for port.
 	 * \param decl connection expression.
 	 */
	void SetConnection( CNode* decl );
	/**
 	 * Get wildcard flag
 	 * \return wildcard flag
 	 */
	bool GetWildcard( void ) { return wildcard; }
	/**
 	 * Set wildcard flag
 	 * \param v value to set flag
 	 */
	void SetWildcard( bool v ) { wildcard = v; }
	/**
 	 * Get dotted attribute.
 	 * \return non-zero if port is dotted.
 	 */
	int IsDotted( void ){ return dotted; };
	/**
 	 * Set port direction
 	 * \param d port direction.
 	 */
	void   Direction( Decl_t d ) { direction = d; }
	/**
 	 * Get port direction.
	 * \return port direction.
	 */
	Decl_t Direction() { return direction; }
	/**
 	 * These should never be called directly.
 	 * \cond
 	 */
	void    PreVisit1( int (*func)(CNode*,void*), void* data );
	void    PostVisit1( void (*func)(CNode*, void*), void* data );
	void    PostSubVisit1( CNode* (*func)(CNode*, void*), void* data );
protected:
	/**
 	 * these should never be called.
 	 */
	virtual INT32 GetWidth( void ) { MASSERT( FALSE ); }
	virtual CNode* GetWidthExp( void ) { MASSERT( FALSE ); } 
	/**
 	 * This should never be call directly
 	 */
        void Copy( CObstack* heap, const CPort& port );
private:
	/*
 	 * Disable copy constructor.
 	 */
	CPort( const CPort& port );
	/**
 	 * \endcond
 	 */
};

#endif // CPORT_HPP
