/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cportdir.hpp
 *		- class definition of port direction
 *		  definition nodes
 *
 ******************************************************************************
 */

#ifndef CPORTDIR_HPP
#define CPORTDIR_HPP

#include <stdio.h>
#include "glue.h"
#include "cnode.h"
#include "cdecl.h"


class CNode;

/**
 * Declaration object for input/output/inout statements.
 */
class CPortDir : public CDecl
{
private:
	CDecl*	decl;		///< variable declaration association
	bool    hasExpression;  ///< port dir has an expression (used by modport)
	CNode*  expression;     ///< port dir expression (used by modport)
public:
	/**
 	 * Create a port direction declaration.
 	 * \param symbol name of declaration.
	 * \param aLoc file coordinates of declaration.
	 * \param type port declaration type.
	 * \param dataType port data type
	 */
	CPortDir( CSymbol* symbol, Coord_t* aLoc, 
		  Decl_t type, CDataType* dataType );
	/**
 	 * Create a clone of this declaration.
 	 * \param heap heap to use for allocation.
 	 * \return new declaration.
 	 */
	virtual CDecl* Clone( CObstack* heap );
	/**
 	 * Dump declaration to file descriptor.
	 * \param f file descriptor.
	 */
	virtual void Dump( FILE* f );
	/**
 	 * Get variable declaration associated with this.
 	 * \return variable declaration.
	 */
	CDecl* GetDecl( void ) { return decl; }
	/**
 	 * Set association to variable declaration.
 	 * \param d variable declaration.
 	 */
	void SetDecl( CDecl* d ) { decl = d; }
	/**
 	 * Get expression
 	 * \return expression
	 */
	CNode* GetExpression( void ) { return expression; }
	/**
 	 * Set expression
 	 * \param e expression
 	 */
	void SetExpression( CNode* e ) { expression = e; }
	/**
 	 * Get has expression flag
 	 * \return has expression flag
	 */
	bool GetHasExpression( void ) { return hasExpression; }
	/**
 	 * Set has expression flag
 	 * \param f has expression flag
 	 */
	void SetHasExpression( bool f ) { hasExpression = f; }
protected:
	/**
 	 * Deep copy of declaration.
 	 * \param heap heap to use for allocation.
 	 * \param d declaration to copy.
 	 */
	void Copy( CObstack* heap, CPortDir& d );
private:
	/*
 	 * Disable copy constructor.
 	 */
	CPortDir( const CPortDir& );
public:
	virtual void PreVisit1( int (*func)(CNode*,void*), void* data );
	virtual void PostVisit1( void (*func)(CNode*, void*), void* data );
	virtual void PostSubVisit1( CNode* (*func)(CNode*, void*), void* data );
};

#endif // CPORT_HPP
