/*****************************************************************************
 * Copyright (C) 1997-2010, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cvar.cpp
 *		- class definition of variable 
 *		  definition nodes
 *
 ******************************************************************************
 */

#include "cvar.h"
#include "cnode.h"
#include "cvector.h"
#include "cdatatype.h"




/****************************************************
	Constructor
*****************************************************/
CVar::CVar( CSymbol* name, Coord_t* aLoc, CDataType* dataType, int undefined ) 
    : CDecl( name, aLoc, eVAR, dataType, 
	     Or(eFLAG_PRAGMA,eFLAG_ARRAY,eFLAG_DATATYPE) ), 
      undefined(undefined) {
}

/****************************************************
	Clone
	- Create a new declaration that is a deep
	  of this declaration.
*****************************************************/
CDecl* CVar::Clone( CObstack* heap )
{
    CVar* clone = new(heap) CVar( GetSymbol(), GetCoord(), NULL, undefined );
    clone->Copy( heap, *this );
    return clone;
}

/****************************************************
	Copy 
	- perform a deep copy
*****************************************************/
void CVar::Copy( CObstack* heap, CVar& var )
{
    CDecl::Copy( heap, var );
    undefined = var.undefined;
}

/****************************************************
	Dump
*****************************************************/
void	CVar::Dump( FILE* f )
{
    fprintf( f, "%s ", declName[GetType()] );
    CDataType* dataType = GetDataType();
    if( dataType ) {
	dataType->Dump( f );
    }
    fprintf( f, ": %s, defined in ", GetName() ); 
    CDecl::Dump( f );
    fprintf( f, "\n" );
}	

