/*****************************************************************************
 * Copyright (C) 1997-2010, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   cvar.hpp
 *		- class definition of variable
 *		  definition nodes
 *
 ******************************************************************************
 */

#ifndef CVAR_H
#define CVAR_H

#include <stdio.h>
#include "glue.h"
#include "cnode.h"
#include "cdecl.h"
#include "cdatatype.h"


class CNode;

/*
 * handle variable declarations
 */

/**
 * Declaration object for variables.
 */
class CVar: public CDecl
{
private:
	int	undefined;	// TRUE if undefined in source
public:
	/**
 	 * Create a register declaration.
 	 * \param symbol declaration symbol.
 	 * \param aLoc file coordinates.
 	 * \param dataType variable data type.
 	 * \param undefined non-zero if register is undefined in source.
 	 */
	CVar( CSymbol* symbol, Coord_t* aLoc, 
	      CDataType* dataType, int undefined = FALSE );
	/**
 	 * Create a clone of this declaration.
 	 * \param heap heap to use for allocation.
 	 * \return new declaration.
 	 */
	virtual CDecl* Clone( CObstack* heap );
	/**
 	 * Get undefined attribute.
 	 * \return undefined attribute.
 	 */
	int	GetUndefined( void ) { return undefined; }
	/**
 	 * Set undefined attribute.
 	 * \param value set undefined attribute.
 	 */
	void	SetUndefined( int value ) { undefined = value; }
	/**
 	 * Dump register info to file.
 	 * \param f file descriptor.
 	 */
	virtual void Dump( FILE* f );
protected:
	/*
 	 * Deep Copy.
 	 */
	void Copy( CObstack* heap, CVar& var );
private:
	/*
 	 * Disable copy constructor.
 	 */
	CVar( const CVar& var );
};

#endif // CVAR_H
