/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   external.hpp
 *		- manager for external ref info 
 *
 ******************************************************************************
 */

#ifndef EXTERNAL_HPP
#define EXTERNAL_HPP

/**
 * This class is a stub that will be used to create
 * linkages for external references.
 * \todo Implement external linkages.
 */
class External {
public:
	/**
 	 * Initialize class info.
 	 */
	static void Initialize() {}
	/**
 	 * Get width of external variable.
 	 * \param symbol external path.
 	 * \return expression width
 	 */
	static int Width( CSymbol* symbol ) { return 32; }
	/**
 	 * Get width of external variable as an expression.
 	 * \param symbol external path.
 	 * \return width expression
 	 */
	static CNode* WidthExp( CSymbol* symbol ) { return NULL; }
	/**
 	 * Determine if width of external variable is constant.
 	 * \param symbol external path.
 	 * \return non-zero if width is constant. ie dependent upon only
 	 *         on parameters and constants.
 	 */
	static int WidthConstant( CSymbol* symbol ) { return FALSE; }
	/**
 	 * Determine if width of external variable is volatile.
 	 * \param symbol external path.
 	 * \return non-zero if width is volatile. ie dependent upon only
 	 *         on parameters or variables.
 	 */
	static int WidthVolatile( CSymbol* symbol ) { return TRUE; }
	/**
 	 * Determine if width of external variable can be evaluated.
 	 * \param symbol external path.
 	 * \return non-zero if width can be evaluated.
 	 */
	static int WidthEvaluateable( CSymbol* symbol ) { return FALSE; }
	/**
 	 * Determine type of external variable.
 	 * \param symbol external path.
 	 * \return type of external variable.
 	 */
	static NodeType_t Type( CSymbol* symbol ) { return eB; }
};

#endif // EXTERNAL_HPP
