/*****************************************************************************
 * Copyright (C) 1997-2009, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   member.h
 *		- member info routines
 *
 ******************************************************************************
 */

#ifndef MEMBER_H
#define MEMBER_H

#include "cnode_def.h"
#include "cdecl.h"
#include "cgenvar.h"
#include "cdatatype.h"
#include "cstructtype.h"
#include "cproxytype.h"

/**
 * Class method to query array info.
 */
class Member {
public:
	/**
 	 * Get width of structure member.
 	 * \param ref reference to a decl
 	 * \param index index list
 	 * \return width of expression
 	 */
	static int Width( CNode* ref, CSymbol* member );
	/**
 	 * Get width of structure member as an expression.
 	 * \param ref reference to a decl
 	 * \param index index list
 	 * \return width expression.
 	 */
	static CNode* WidthExp( CNode* ref, CSymbol* member );
	/**
 	 * Determine if width of structure member is constant.
 	 * \todo current not fully implemented.
 	 * \param ref reference to a decl
 	 * \param index index list
 	 * \return non-zero if width is constant. ie dependent upon only
 	 *         on parameters and constants.
 	 */
	static int WidthConstant( CNode* ref, CSymbol* member );
	/**
 	 * Determine if width of structure member is volatile.
 	 * \todo current not fully implemented.
 	 * \param ref reference to a decl
 	 * \param index index list
 	 * \return non-zero if width is volatile. ie dependent upon only
 	 *         on parameters or variables.
 	 */
	static int WidthVolatile( CNode* ref, CSymbol* member );
	/**
 	 * Determine if width of structure member can be evaluated.
 	 * \todo current not fully implemented.
 	 * \param ref reference to a decl
 	 * \param index index list
 	 * \return non-zero if width can be evaluated.
 	 */
	static int WidthEvaluateable( CNode* ref, CSymbol* member ) ;
	/**
 	 * Determine type of structure member.
 	 * \todo current not fully implemented.
 	 * \param ref reference to a decl
 	 * \param index index list
 	 * \return type of array return value.
 	 */
	static NodeType_t Type( CNode* ref, CSymbol* member );
	/**
	 * Get datatype info for struct reference
	 * \param ref reference to a decl/subreference
	 * \param index index list 
	 * \param dataType resolved base data type
	 * \param unpackedDimensions list of unresolved unpacked dimensions
	 * \return -1 if reference is valid but unresolvable
	 * 	    0 if reference is invalid
	 * 	    1 if reference is valid and was resolved
	 */
	static int ResolveReference( 
		CNode*      ref, 
		CSymbol*    symbol,
		CDataType*& dataType,
		vector<CNode*>& unpackedDimensions,
		vector<CNode*>& packedDimensions );
};

#endif // MEMBER_H
