/*****************************************************************************
 * Copyright (C) 1997-2007, Mark Hummel
 * This file is part of Vrq.
 *
 * Vrq is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * Vrq is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 *****************************************************************************
 */
/******************************************************************************
 *
 *
 *	   systask.hpp
 *		- manager for systask info 
 *
 ******************************************************************************
 */

#ifndef SYSTASK_HPP
#define SYSTASK_HPP

/**
 * Class method to query systask info.
 */
class Systask {
public:
	/**
 	 * Initialize class info.
 	 */
	static void Initialize() {}
	/**
 	 * Get width of systask.
 	 * \param symbol systask name.
 	 * \param args systask arguments.
 	 * \return width of expression.
 	 */
	static int Width( CSymbol* symbol, CNode* args ) 
	{ 
	    /*
 	     * implement builtin functions
 	     */
	    if( !strcmp(symbol->GetName(), "$signed" ) ) {
	 	return args->GetWidth();
	    }
	    return 32; 
	}
	/**
 	 * Get width of systask as an expression.
 	 * \param symbol systask name.
 	 * \param args systask arguments.
 	 * \return width expression.
 	 */
	static CNode* WidthExp( CSymbol* symbol, CNode* args ) 
	{ 
	    /*
 	     * implement builtin functions
 	     */
	    if( !strcmp(symbol->GetName(), "$signed" ) ) {
	 	return args->GetWidthExp();
	    }
	    return cINT32(32); 
	}
	/**
 	 * Determine if width of systask variable is constant.
 	 * \todo current not fully implemented.
 	 * \param symbol systask name.
 	 * \param args systask arguments.
 	 * \return non-zero if width is constant. ie dependent upon only
 	 *         on parameters and constants.
 	 */
	static int WidthConstant( CSymbol* symbol, CNode* args ) 
	{ 
	    /*
 	     * implement builtin functions
 	     */
	    if( !strcmp(symbol->GetName(), "$signed" ) ) {
	 	return args->IsWidthConstant();
	    }
	    return FALSE; 
	}
	/**
 	 * Determine if width of systask variable is volatile.
 	 * \todo current not fully implemented.
 	 * \param symbol systask name.
 	 * \param args systask arguments.
 	 * \return non-zero if width is volatile. ie dependent upon only
 	 *         on parameters or variables.
 	 */
	static int WidthVolatile( CSymbol* symbol, CNode* args ) 
	{ 
	    /*
 	     * implement builtin functions
 	     */
	    if( !strcmp(symbol->GetName(), "$signed" ) ) {
	 	return args->IsWidthVolatile();
	    }
	    return TRUE; 
	}
	/**
 	 * Determine if width of systask can be evaluated.
 	 * \todo current not fully implemented.
 	 * \param symbol systask name.
 	 * \param args systask arguments.
 	 * \return non-zero if width can be evaluated.
 	 */
	static int WidthEvaluateable( CSymbol* symbol, CNode* args ) 
	{ 
	    /*
 	     * implement builtin functions
 	     */
	    if( !strcmp(symbol->GetName(), "$signed" ) ) {
	 	return args->IsWidthEvaluateable();
	    }
	    return FALSE; 
	}
	/**
 	 * Determine type of systask.
 	 * \todo current not fully implemented.
 	 * \param symbol systask name.
 	 * \param args systask arguments.
 	 * \return type of systask return value.
 	 */
	static NodeType_t Type( CSymbol* symbol, CNode* args ) 
	{ 
	    /*
 	     * implement builtin functions
 	     */
	    if( !strcmp(symbol->GetName(), "$signed" ) ) {
	 	return eS;
	    }
	    return eB; 
	}
};

#endif // SYSTASK_HPP
